# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging, re

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.experiment_artifact import ExperimentArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.api_client import ApiException
import json
from .wml_experiment_adapter import WmlExperimentCollectionAdapter
from ibm_watson_machine_learning.libs.repo.swagger_client.models import ExperimentInput,TagRepository,\
    ExperimentInputSettings,MetricObjectExperiments,HyperParametersOptimizationExperiments,EvaluationDefinitionExperiments
from ibm_watson_machine_learning.libs.repo.swagger_client.models import TrainingReferenceExperiments,ComputeConfigurationExperiments
from ibm_watson_machine_learning.libs.repo.swagger_client.models import  ConnectionObjectTargetExperiments,ConnectionObjectSourceExperiments,AuthorExperiments, PatchOperationExperiments, HyperParametersExperiments
from ibm_watson_machine_learning.libs.repo.swagger_client.models import HyperParametersOptimizationExperimentsMethodParameters, HyperParametersOptimizationExperimentsMethod, \
    HyperParametersExperimentsDoubleRange, HyperParametersExperimentsIntRange

logger = logging.getLogger('WmlExperimentCollection')


class WmlExperimentCollection:
    """
    Client operating on experiments in repository service.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLRepositoryApi repository_api: client connecting to repository rest api
    :param MLRepositoryClient client: high level client used for simplification and argument for constructors
    """
    def __init__(self, base_path, repository_api, client):

        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryApi
        if not isinstance(base_path, str) and not isinstance(base_path, unicode):
            raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

        if not isinstance(repository_api, MLRepositoryApi):
            raise ValueError('Invalid type for repository_api: {}'.format(repository_api.__class__.__name__))

        if not isinstance(client, MLRepositoryClient):
            raise ValueError('Invalid type for client: {}'.format(client.__class__.__name__))

        self.base_path = base_path
        self.repository_api = repository_api
        self.client = client


    def all(self, queryMap=None):
        """
        Gets info about all experiments which belong to this user.

        Not complete information is provided by all(). To get detailed information about experiment use get().

        :return: info about experiments
        :rtype: list[ExperimentsArtifact]
        """
        all_experiments = self.repository_api.repository_listexperiments(queryMap)
        list_experiment_artifact = []
        if all_experiments is not None:
            resr = all_experiments.resources

            for iter1 in resr:
                list_experiment_artifact.append(WmlExperimentCollectionAdapter(iter1, self.client).artifact())
            return list_experiment_artifact
        else:
            return []


    def get(self, experiment):

        """
        Gets detailed information about experiment.

        :param str experiment_id: uid used to identify experiment
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(experiment, str) and  not isinstance(experiment, unicode):
            raise ValueError('Invalid type for experiment_id: {}'.format(experiment.__class__.__name__))
        if(experiment.__contains__("/v3/experiments")):
            matched = re.search('.*/v3/experiments/([A-Za-z0-9\-]+)', experiment)
            if matched is not None:
                experiment_id = matched.group(1)
                return self.get(experiment_id)
            else:
                raise ValueError('Unexpected artifact href: {} format'.format(experiment))
        else:
            experiment_output = self.repository_api.v3_experiments_id_get(experiment)
            if experiment_output is not None:
                return WmlExperimentCollectionAdapter(experiment_output,self.client).artifact()
            else:
                raise Exception('Experiment not found'.format(experiment))


    def remove(self, experiment):
        """
        Removes experiment with given experiment_id.

        :param str experiment_id: uid used to identify experiment
        """

        if not isinstance(experiment, str) and not isinstance(experiment, unicode):
            raise ValueError('Invalid type for experiment_id: {}'.format(experiment.__class__.__name__))
        if(experiment.__contains__("/v3/experiments")):
            matched = re.search('.*/v3/experiments/([A-Za-z0-9\-]+)', experiment)
            if matched is not None:
                experiment_id = matched.group(1)
                self.remove(experiment_id)
            else:
                raise ValueError('Unexpected experiment artifact href: {} format'.format(experiment))
        else:
            return self.repository_api.v3_experiments_id_delete(experiment)

    def patch(self, experiment_id, artifact):
        experiment_patch_input = self.prepare_experiment_patch_input(artifact)
        experiment_patch_output = self.repository_api.v3_experiments_id_patch_with_http_info(experiment_id, experiment_patch_input)
        statuscode = experiment_patch_output[1]

        if statuscode != 200:
            logger.info('Error while patching experiment: no location header')
            raise ApiException(statuscode,"Error while patching experiment")

        if experiment_patch_output is not None:
            new_artifact =  WmlExperimentCollectionAdapter(experiment_patch_output[0],self.client).artifact()
        return new_artifact

    def save(self, artifact):
        """
        Saves experiment in repository service.

        :param SparkPipelineArtifact artifact: artifact to be saved in the repository service
        :return: saved artifact with changed MetaProps
        :rtype: SparkPipelineArtifact
        """
        logger.debug('Creating a new WML experiment: {}'.format(artifact.name))

        if not issubclass(type(artifact), ExperimentArtifact):
            raise ValueError('Invalid type for artifact: {}'.format(artifact.__class__.__name__))

        experiment_input = self._prepare_wml_experiment_input(artifact)
        experiment_output = self.repository_api.wml_assets_experiment_creation_with_http_info(experiment_input)


        statuscode = experiment_output[1]
        if statuscode != 201:
            logger.info('Error while creating experiment: no location header')
            raise ApiException(statuscode, 'No artifact location')

        if experiment_output is not None:
            new_artifact =  WmlExperimentCollectionAdapter(experiment_output[0],self.client).artifact()
        return new_artifact



    @staticmethod
    def _prepare_wml_experiment_input(artifact):

        tags_data_list = None
        settings_data = None
        training_reference_list=None
        training_data_ref = None
        training_results_ref=None
        hyper_param_list = None

        #tags
        tags=artifact.meta.prop(MetaNames.EXPERIMENTS.TAGS)
        if isinstance(tags, str):
            tags_list = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.TAGS))
            tags_data_list = []
            if isinstance(tags_list, list):
               for iter1 in tags_list:
                 tags_data = TagRepository()
                 for key in iter1:
                    if key == 'value':
                        tags_data.value= iter1['value']
                    if key == 'description':
                        tags_data.description = iter1['description']
                 tags_data_list.append(tags_data)
            else:
                raise ValueError("Invalid tag Input")

        #settings
        if artifact.meta.prop(MetaNames.EXPERIMENTS.SETTINGS) is  None:
           raise ValueError("MetaNames.EXPERIMENTS.SETTINGS not defined")

        settings=artifact.meta.prop(MetaNames.EXPERIMENTS.SETTINGS)
        if isinstance(settings, str):
            settings = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.SETTINGS))

            if isinstance(settings, dict):
                author=settings.get('author', None)
                author_experiment = None
                if author is not None:
                    author_name = author.get('name',None)
                    author_experiment = AuthorExperiments(author_name)


                evaluation_definition = settings.get('evaluation_definition',None)
                evaluation_definition_exp = None
                if evaluation_definition is not None:
                    evaluation_definition_method = evaluation_definition.get('method',None)
                    evaluation_definition_metrics = evaluation_definition.get('metrics')
                    metrics_experiments = []
                    if isinstance(evaluation_definition_metrics, list):
                      for iter1 in evaluation_definition_metrics:
                        metrics_data = MetricObjectExperiments()
                        for key in iter1:
                            if key == 'name':
                                metrics_data.name= iter1['name']
                            if key == 'maximize':
                                metrics_data.maximize= iter1['maximize']
                            metrics_experiments.append(metrics_data)
                    else:
                        raise ValueError("Invalid Input: Metrics list Expected")
                    evaluation_definition_exp = EvaluationDefinitionExperiments(evaluation_definition_method,metrics_experiments)


                settings_data = ExperimentInputSettings(
                    name = settings.get('name'),
                    description = settings.get('description', None),
                    author=author_experiment,
                    label_column= settings.get('label_column', None),
                    evaluation_definition= evaluation_definition_exp
                    )

        #training_refrences
        training_ref=artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_REFERENCES)
        if isinstance(training_ref, str):
            training_ref_list = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_REFERENCES))
            training_reference_list = []
            if isinstance(training_ref_list, list):
                for iter1 in training_ref_list:
                    compute = iter1.get('compute_configuration', None)
                    compute_config_object = None
                    if compute is not None:
                        compute_name = compute.get('name')
                        compute_nodes = compute.get('nodes',None)
                        compute_config_object = ComputeConfigurationExperiments(compute_name, compute_nodes)

                    hyper_params_optimization = iter1.get('hyper_parameters_optimization', None)
                    hyper_parameters_optimization_object = None

                    if hyper_params_optimization is not None:

                        hyper_parameters_optimization_method = hyper_params_optimization.get('method')
                        method_name = hyper_parameters_optimization_method.get('name', None)

                        parameters = hyper_parameters_optimization_method.get('parameters', None)
                        params_list = []
                        if parameters is not None:

                            if isinstance(parameters, list):

                                for iteration in parameters:
                                    param_name = None
                                    param_string_value = None
                                    param_double_value = None
                                    param_int_value = None
                                    for key in iteration:

                                        if(key == 'name'):
                                            param_name = iteration['name']
                                        if(key == 'string_value'):
                                            param_string_value = iteration['string_value']
                                        if(key == 'double_value'):
                                            param_double_value = iteration['double_value']
                                        if(key == 'int_value'):
                                           param_int_value = iteration['int_value']
                                    parameters_object = HyperParametersOptimizationExperimentsMethodParameters(param_name, param_string_value, param_double_value, param_int_value)
                                    params_list.append(parameters_object)
                        hyper_parameters_optimization_method_object = HyperParametersOptimizationExperimentsMethod(method_name, params_list)

                        hyper_params = hyper_params_optimization.get('hyper_parameters', None)
                        if hyper_params is not None:
                            if isinstance(hyper_params, list):
                                hyper_param_list = []
                                double_value_list = []
                                int_value_list = []
                                string_value_list = []
                                double_range_object = None
                                int_range_object = None
                                for iter2 in hyper_params:
                                    for key in iter2:

                                        if key == 'name':

                                            hyper_params_name= iter2['name']
                                        if key == 'double_values':
                                            double_values = iter2['double_values']
                                            if double_values is not None:
                                                if isinstance(double_values, list):

                                                    for double_value in double_values:

                                                        double_value_list.append(double_value)

                                        if key == 'int_values':
                                            int_values = iter2['int_values']
                                            if int_values is not None:
                                                if isinstance(int_values, list):

                                                    for int_value in int_values:

                                                        int_value_list.append(int_value)

                                        if key == 'string_values':

                                            string_values = iter2['string_values']
                                            if string_values is not None:
                                                if isinstance(string_values, list):

                                                    for value in string_values:

                                                        string_value_list.append(value)

                                        if key == 'double_range':
                                            double_values_range = iter2['double_range']
                                            if double_values_range is not None:
                                                double_min_value = double_values_range.get('min_value')
                                                double_max_value = double_values_range.get('max_value')
                                                double_step = double_values_range.get('step', None)
                                                double_power = double_values_range.get('power', None)
                                                double_range_object = HyperParametersExperimentsDoubleRange(
                                                    double_min_value, double_max_value, double_step, double_power)

                                        if key == 'int_range':


                                            int_values_range = iter2['int_range']
                                            if int_values_range is not None:
                                                int_min_value = int_values_range.get('min_value')
                                                int_max_value = int_values_range.get('max_value')
                                                int_step = int_values_range.get('step', None)
                                                int_power = int_values_range.get('power', None)
                                                int_range_object = HyperParametersExperimentsIntRange(
                                                    int_min_value, int_max_value, int_step, int_power)

                                    if not double_value_list:
                                        double_value_list=None
                                    if not int_value_list:
                                        int_value_list=None

                                    hyper_params_object = HyperParametersExperiments(hyper_params_name,
                                                                                     double_value_list,
                                                                                     int_value_list,
                                                                                     string_value_list,
                                                                                     double_range_object,
                                                                                     int_range_object)
                                    hyper_param_list.append(hyper_params_object)
                        hyper_parameters_optimization_object = HyperParametersOptimizationExperiments(hyper_parameters_optimization_method_object, hyper_param_list)

                    training_reference = TrainingReferenceExperiments(
                            name = iter1.get('name'),
                            training_definition_url= iter1.get('training_definition_url'),
                            command = iter1.get('command', None),
                            hyper_parameters_optimization = hyper_parameters_optimization_object,
                            compute_configuration = compute_config_object,
                            pretrained_model_url = iter1.get('pretrained_model_url', None)
                        )
                    training_reference_list.append(training_reference)

            else:
                raise ApiException(404, 'Invalid Input')

        #Training_data_reference
        if artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE) is  None:
            raise ValueError("MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE not defined")

        training_data_ref = artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE)
        if isinstance(training_data_ref, str):
            dataref = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE))
            if isinstance(dataref, dict):
                training_data_ref = ConnectionObjectSourceExperiments(
                        dataref.get('type'),
                        dataref.get('connection'),
                        dataref.get('source')
                    )
            else:
                raise ApiException(404, 'Invalid  TRAINING_DATA_REFERENCE Input')

        #Training_result_refernce
        if artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_RESULTS_REFERENCE) is not None:
            training_results_ref = artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_RESULTS_REFERENCE)
            if isinstance(training_results_ref, str):
                resultref_dict = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.TRAINING_RESULTS_REFERENCE))
                if isinstance(resultref_dict, dict):
                    training_results_ref = ConnectionObjectTargetExperiments(
                        resultref_dict.get('type', None),
                        resultref_dict.get('connection', None),
                        resultref_dict.get('target', None)
                    )
                else:
                    raise ApiException(404, 'Invalid TRAINING_RESULTS_REFERENCE Input')


        experiments_input = ExperimentInput(
            tags=tags_data_list,
            settings=settings_data,
            training_references=training_reference_list,
            training_data_reference=training_data_ref,
            training_results_reference=training_results_ref

        )

        return experiments_input


    @staticmethod
    def prepare_experiment_patch_input(artifact):
        patch_list =[]
        patch_input = artifact.meta.prop(MetaNames.EXPERIMENTS.PATCH_INPUT)
        if isinstance(patch_input, str):
            patch_input_list = json.loads(artifact.meta.prop(MetaNames.EXPERIMENTS.PATCH_INPUT))
            if isinstance(patch_input_list, list):
                for iter1 in patch_input_list:
                    experiment_patch = PatchOperationExperiments(
                        op = iter1.get('op'),
                        path= iter1.get('path'),
                        value = iter1.get('value', None),
                        _from =iter1.get('from', None),
                    )
                    patch_list.append(experiment_patch)

                return patch_list
            else:
                raise ApiException(404, 'Invalid Patch Input')
        else:
            raise ApiException(404, 'Invalid Patch Input')



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import requests
import json
from ibm_watson_machine_learning.utils import DEPLOYMENT_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, print_text_header_h1, print_text_header_h2, STR_TYPE, STR_TYPE_NAME, docstring_parameter, str_type_conv, \
    StatusLogger, meta_props_str_conv, convert_metadata_to_parameters
from ibm_watson_machine_learning.wml_client_error import WMLClientError, MissingValue, NoVirtualDeploymentSupportedForICP
from ibm_watson_machine_learning.href_definitions import is_uid
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.metanames import DeploymentMetaNames, ScoringMetaNames, DecisionOptimizationMetaNames, DeploymentNewMetaNames
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
import numpy as np
#from ibm_watson_machine_learning.metanames import DeploymentMetaNames, ScoreMetaNames, EnvironmentMetaNames

lib_checker = LibraryChecker()

class Deployments(WMLResource):
    """
        Deploy and score published artifacts (models and functions).

    """
    cloud_platform_spaces = False
    icp_platform_spaces = False

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        if not client.ICP and not client.CLOUD_PLATFORM_SPACES and not client.ICP_PLATFORM_SPACES:
            Deployments._validate_type(client.service_instance.details, u'instance_details', dict, True)
            Deployments._validate_type_of_details(client.service_instance.details, INSTANCE_DETAILS_TYPE)
        self.session = requests.Session()
        self._ICP = client.ICP
        if client.CLOUD_PLATFORM_SPACES or client.ICP_PLATFORM_SPACES:
            self.ConfigurationMetaNames = DeploymentNewMetaNames()
        else:
            self.ConfigurationMetaNames = DeploymentMetaNames()

        if client.CLOUD_PLATFORM_SPACES:
            Deployments.cloud_platform_spaces = True

        if client.ICP_PLATFORM_SPACES:
            Deployments.icp_platform_spaces = True

        self.ScoringMetaNames = ScoringMetaNames()
        self.DecisionOptimizationMetaNames = DecisionOptimizationMetaNames()

    def _deployment_status_errors_handling(self, deployment_details, operation_name, deployment_id):
        try:
            if 'failure' in deployment_details['entity']['status']:
                errors = deployment_details[u'entity'][u'status'][u'failure'][u'errors']
                for error in errors:
                    if type(error) == str:
                        try:
                            error_obj = json.loads(error)
                            print(error_obj[u'message'])
                        except:
                            print(error)
                    elif type(error) == dict:
                        print(error['message'])
                    else:
                        print(error)
                raise WMLClientError('Deployment ' + operation_name + ' failed for deployment id: ' + deployment_id +'. Errors: ' + str(errors))
            else:
                print(deployment_details['entity']['status'])
                raise WMLClientError('Deployment ' + operation_name + ' failed for deployment id: ' + deployment_id +'. Error: ' + str(deployment_details['entity']['status']['state']))
        except WMLClientError as e:
            raise e
        except Exception as e:
            self._logger.debug('Deployment ' + operation_name + ' failed: ' + str(e))
            print(deployment_details['entity']['status']['failure'])
            raise WMLClientError('Deployment ' + operation_name + ' failed for deployment id: ' + deployment_id + '.')

    @docstring_parameter({'str_type': STR_TYPE_NAME})  # TODO model_uid and artifact_uid should be changed to artifact_uid only
    def create(self, artifact_uid=None, meta_props=None, rev_id=None, **kwargs):
            """
                Create a deployment from an artifact. As artifact we understand model or function which may be deployed.

                **Parameters**

                .. important::

                    #. **artifact_uid**:  Published artifact UID (model or function uid)\n
                       **type**: str\n

                    #. **meta_props**: metaprops. To see the available list of metanames use:\n
                                       >>> client.deployments.ConfigurationMetaNames.get()

                       **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of the created deployment\n
                    **return type**: dict\n

                **Example**

                 >>> meta_props = {
                 >>> wml_client.deployments.ConfigurationMetaNames.NAME: "SAMPLE DEPLOYMENT NAME",
                 >>> wml_client.deployments.ConfigurationMetaNames.ONLINE: {},
                 >>> wml_client.deployments.ConfigurationMetaNames.HARDWARE_SPEC : { "id":  "e7ed1d6c-2e89-42d7-aed5-8sb972c1d2b"}
                 >>> }
                 >>> deployment_details = client.deployments.create(artifact_uid, meta_props)

             """
            #artifact_uid = str_type_conv(artifact_uid) if artifact_uid is not None else (str_type_conv(kwargs['model_uid'] if 'model_uid' in kwargs else None))
            ##To be removed once deployments adds support for projects
            WMLResource._chk_and_block_create_update_for_python36(self)
            self._client._check_if_space_is_set()

            Deployments._validate_type(artifact_uid, u'artifact_uid', STR_TYPE, True)

            if self._ICP:
                predictionUrl = self._wml_credentials[u'url']

            if meta_props is None:
                raise WMLClientError("Invalid input. meta_props can not be empty.")

            if self._client.CLOUD_PLATFORM_SPACES and 'r_shiny' in meta_props:
                raise WMLClientError('Shiny is not supported in this release')

            metaProps = self.ConfigurationMetaNames._generate_resource_metadata(meta_props)

            if 'serving_name' in str(metaProps) and 'r_shiny' in str(metaProps):
                metaProps['serving_name'] = meta_props['serving_name']
                del metaProps['online']

            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'wml_instance_id' in meta_props:
                metaProps.update({'wml_instance_id': meta_props['wml_instance_id']})

            if not self._client.ICP_30 and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                ## Remove new meataprop from older one, just in case if user provides.
                if self.ConfigurationMetaNames.HARDWARE_SPEC in metaProps:
                    metaProps.pop(self.ConfigurationMetaNames.HARDWARE_SPEC)
                if self.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS in metaProps:
                    metaProps.pop(self.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS)
                if self.ConfigurationMetaNames.R_SHINY in metaProps:
                    metaProps.pop(self.ConfigurationMetaNames.R_SHINY)

                artifact_details = self._client.repository.get_details(artifact_uid)
                artifact_href = artifact_details['metadata']['href']
                artifact_revision_id = artifact_href.split('=')[1]

                if 'model' in artifact_href:
                    if self._client.CAMS:
                        details = "/v4/models/"+artifact_uid + "?space_id=" + self._client.default_space_id
                    else:
                        details = "/v4/models/" + artifact_uid  + "?rev=" + artifact_revision_id
                elif 'function' in artifact_href:
                    if self._client.CAMS:
                        details = "/v4/functions/" + artifact_uid + "?space_id=" + self._client.default_space_id
                    else:
                        details = "/v4/functions/" + artifact_uid  + "?rev=" + artifact_revision_id
                else:
                    raise WMLClientError('Unexpected artifact type: \'{}\'.'.format(artifact_uid))
                if 'model' in artifact_href:
                    if "tensorflow_1.11" in artifact_details["entity"]["type"] or "tensorflow_1.5" in \
                            artifact_details["entity"]["type"]:
                        print("Note: Model of framework tensorflow and versions 1.5/1.11 has been deprecated. These versions will not be supported after 26th Nov 2019.")

                metaProps['asset'] = {'href': details}
                if self._client.CAMS:
                    if self._client.default_space_id is not None:
                        metaProps['space'] = {'href': "/v4/spaces/" + self._client.default_space_id}
                    else:
                        raise WMLClientError(
                            "It is mandatory is set the space. Use client.set.default_space(<SPACE_GUID>) to set the space.")

            ##Check if default space is set
            else:
                metaProps['asset'] = metaProps.get('asset') if metaProps.get('asset') else {'id': artifact_uid}
                if rev_id is not None:
                    metaProps['asset'].update({'rev': rev_id})

                if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    metaProps['space_id'] = self._client.default_space_id
                else:
                    if 'space' not in metaProps:
                        if self._client.default_space_id is not None:
                            metaProps['space_id'] = self._client.default_space_id
                        else:
                            raise WMLClientError(
                                "It is mandatory is set the space. Use client.set.default_space(<SPACE_GUID>) to set the space.")

            url = self._client.service_instance._href_definitions.get_deployments_href()
            if not self._ICP:
                if self._client.CLOUD_PLATFORM_SPACES:
                    response = requests.post(
                        url,
                        json=metaProps,
                        params=self._client._params(),  # version is mandatory
                        headers=self._client._get_headers())
                else:
                    response = requests.post(
                        url,
                        json=metaProps,
                        headers=self._client._get_headers())
            else:
                if self._client.ICP_PLATFORM_SPACES:
                    response = requests.post(
                        url,
                        json=metaProps,
                        params=self._client._params(),  # version is mandatory
                        headers=self._client._get_headers(),
                        verify=False)
                else:
                    response = requests.post(
                        url,
                        json=metaProps,
                        headers=self._client._get_headers(),
                        verify=False)

            ## Post Deployment call executed
            if response.status_code == 202:
                deployment_details = response.json()
                #print(deployment_details)

                if self._ICP:
                    if 'online_url' in deployment_details["entity"]["status"]:
                     scoringUrl = deployment_details.get(u'entity').get(u'status').get('online_url').get('url').replace('https://ibm-nginx-svc:443', predictionUrl)
                     deployment_details[u'entity'][u'status']['online_url']['url'] = scoringUrl

                deployment_uid = self.get_uid(deployment_details)

                import time
                print_text_header_h1(u'Synchronous deployment creation for uid: \'{}\' started'.format(artifact_uid))

                status = deployment_details[u'entity'][u'status']['state']

                notifications = []

                with StatusLogger(status) as status_logger:
                    while True:
                        time.sleep(5)
                        deployment_details = self._client.deployments.get_details(deployment_uid)
                        #this is wrong , needs to update for ICP
                        if "system" in deployment_details:
                            notification = deployment_details['system']['warnings'][0]['message']
                            if notification not in notifications:
                                print("\nNote: " + notification)
                                notifications.append(notification)
                        if self._ICP and not self._client.ICP_PLATFORM_SPACES:
                            scoringUrl = deployment_details.get(u'entity').get(u'asset').get('href').replace('https://wml-os-envoy:16600', predictionUrl)
                            deployment_details[u'entity'][u'asset']['href'] = scoringUrl
                        status = deployment_details[u'entity'][u'status'][u'state']
                        status_logger.log_state(status)
                        if status != u'DEPLOY_IN_PROGRESS' and status != "initializing":
                            break
                if status == u'DEPLOY_SUCCESS' or status == u'ready':
                    print(u'')
                    print_text_header_h2(
                        u'Successfully finished deployment creation, deployment_uid=\'{}\''.format(deployment_uid))
                    return deployment_details
                else:
                    print_text_header_h2(u'Deployment creation failed')
                    self._deployment_status_errors_handling(deployment_details, 'creation', deployment_uid)
            else:
                error_msg = u'Deployment creation failed'
                reason = response.text
                print(reason)
                print_text_header_h2(error_msg)
                raise WMLClientError(error_msg + '. Error: ' + str(response.status_code) + '. ' + reason)

            #return self._handle_response(202, u'created deployment', response)
    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(deployment_details):
        """
            Get deployment_uid from deployment details. Deprecated!! Use get_id(deployment_details) instead

            **Parameters**

            .. important::
                #. **deployment_details**:  Metadata of the deployment\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: deployment UID that is used to manage the deployment\n
                **return type**: str

            **Example**

             >>> deployment_uid = client.deployments.get_uid(deployment)
        """
        Deployments._validate_type(deployment_details, u'deployment_details', dict, True)

        if not Deployments.cloud_platform_spaces and not Deployments.icp_platform_spaces:
            Deployments._validate_type_of_details(deployment_details, DEPLOYMENT_DETAILS_TYPE)

        try:
            if 'id' in deployment_details[u'metadata']:
                uid = deployment_details.get(u'metadata').get(u'id')
            else:
                uid = deployment_details.get(u'metadata').get(u'guid')
        except Exception as e:
            raise WMLClientError(u'Getting deployment UID from deployment details failed.', e)

        if uid is None:
            raise MissingValue(u'deployment_details.metadata.guid')

        return uid

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(deployment_details):
        """
            Get deployment id from deployment details.

            **Parameters**

            .. important::
                #. **deployment_details**:  Metadata of the deployment\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: deployment ID that is used to manage the deployment\n
                **return type**: str

            **Example**

             >>> deployment_id = client.deployments.get_id(deployment)
        """

        return Deployments.get_uid(deployment_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(deployment_details):
        """
            Get deployment_href from deployment details.

            **Parameters**

            .. important::
                #. **deployment_details**:  Metadata of the deployment.\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: Deployment href that is used to manage the deployment.\n
                **return type**: str

            **Example**

            >>> deployment_href = client.deployments.get_href(deployment)

        """
        Deployments._validate_type(deployment_details, u'deployment_details', dict, True)
        if not Deployments.cloud_platform_spaces and not Deployments.icp_platform_spaces:
            Deployments._validate_type_of_details(deployment_details, DEPLOYMENT_DETAILS_TYPE)

        try:
            if 'href' in deployment_details[u'metadata']:
                url = deployment_details.get(u'metadata').get(u'href')
            else:
                url = "/ml/v4/deployments/{}".format(deployment_details[u'metadata'][u'id'])
        except Exception as e:
            raise WMLClientError(u'Getting deployment url from deployment details failed.', e)

        if url is None:
            raise MissingValue(u'deployment_details.metadata.href')

        return url

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, deployment_uid=None, serving_name=None, limit=None, asynchronous=False, get_all=False):
        """
           Get information about your deployment(s). If deployment_uid is not passed, all deployment details are fetched.

           **Parameters**

           .. important::
                #. **deployment_uid**: Unique Id of Deployment (optional)\n
                   **type**: str\n
                #. **serving_name**: Serving name to filter deployments (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                **returns**: metadata of deployment(s)\n
                **return type**: dict\n
                dict (if deployment_uid is not None) or {"resources": [dict]} (if deployment_uid is None)\n

           .. note::
                If deployment_uid is not specified, all deployments metadata is fetched (can be filtered with serving_name)\n


           **Example**

            >>> deployment_details = client.deployments.get_details(deployment_uid)
            >>> deployment_details = client.deployments.get_details(deployment_uid=deployment_uid)
            >>> deployments_details = client.deployments.get_details()
            >>> deployments_details = client.deployments.get_details(limit=100)
            >>> deployments_details = client.deployments.get_details(limit=100, get_all=True)
            >>> deployments_details = []
            >>> for entry in client.deployments.get_details(limit=100, asynchronous=True, get_all=True):
            >>>    deployments_details.extend(entry)

        """
        ##To be removed once deployments adds support for deployments
        self._client._check_if_space_is_set()

        deployment_uid = str_type_conv(deployment_uid)
        Deployments._validate_type(deployment_uid, u'deployment_uid', STR_TYPE, False)

        if deployment_uid is not None and not is_uid(deployment_uid):
            raise WMLClientError(u'\'deployment_uid\' is not an uid: \'{}\''.format(deployment_uid))

        url = self._client.service_instance._href_definitions.get_deployments_href()
        ## TODO:  Adding a work around till v4/deployment accepts space_id query parameter for GET
        if self._ICP and not self._client.ICP_PLATFORM_SPACES:
            if deployment_uid is not None:
                url = url + '/' + deployment_uid
                if self._client.ICP_35 or self._client.ICP_40:
                    response = requests.get(
                        url,
                        params = self._client._params(),
                        headers=self._client._get_headers(),
                        verify=False
                    )
                else:
                    response = requests.get(
                        url,
                        headers=self._client._get_headers(),
                        verify=False
                    )
                return self._handle_response(200, "deployment", response)
            else:

                if asynchronous:
                    return self._get_artifact_details(url, deployment_uid, limit, 'deployments',
                                                      _async=asynchronous, _all=get_all)

                else:
                    deployment_details = self._get_artifact_details(url, deployment_uid, limit, 'deployments',
                                                                    _async=asynchronous, _all=get_all)
                    if "system" in deployment_details:
                        print("Note: " + deployment_details['system']['warnings'][0]['message'])
                    return deployment_details
        else:
            query_params = self._client._params()
            if serving_name:
                query_params['serving_name'] = serving_name

            if deployment_uid is None:
                return self._get_artifact_details(url, deployment_uid, limit, 'deployments', query_params=query_params,
                                                  _async=asynchronous, _all=get_all)

            else:
                return self._get_artifact_details(url, deployment_uid, limit, 'deployments', query_params=query_params)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_scoring_href(deployment_details):
        """
            Get scoring url from deployment details.

            **Parameters**

            .. important::
                #. **deployment_details**:  Metadata of the deployment\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: scoring endpoint url that is used for making scoring requests\n
                **return type**: str\n

            **Example**

             >>> scoring_href = client.deployments.get_scoring_href(deployment)

        """

        Deployments._validate_type(deployment_details, u'deployment', dict, True)

        if not Deployments.cloud_platform_spaces and not Deployments.icp_platform_spaces:
            Deployments._validate_type_of_details(deployment_details, DEPLOYMENT_DETAILS_TYPE)
        scoring_url = None
        try:
            url = deployment_details['entity']['status'].get('online_url')
            if url is not None:
                scoring_url = deployment_details['entity']['status']['online_url']['url']
            else:
                raise MissingValue(u'Getting scoring url for deployment failed. This functionality is  available only for sync deployments')

        except Exception as e:
            raise WMLClientError(u'Getting scoring url for deployment failed. This functionality is  available only for sync deployments', e)

        if scoring_url is None:
            raise MissingValue(u'scoring_url missing in online_predictions')
        return scoring_url

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_serving_href(deployment_details):
        """
            Get serving url from deployment details.

            **Parameters**

            .. important::
                #. **deployment_details**:  Metadata of the deployment\n
                   **type**: dict\n

            **Output**

            .. important::
                **returns**: serving endpoint url that is used for making scoring requests\n
                **return type**: str\n

            **Example**

             >>> scoring_href = client.deployments.get_serving_href(deployment)

        """

        Deployments._validate_type(deployment_details, u'deployment', dict, True)

        if not Deployments.cloud_platform_spaces and not Deployments.icp_platform_spaces:
            Deployments._validate_type_of_details(deployment_details, DEPLOYMENT_DETAILS_TYPE)

        try:
            serving_name = deployment_details['entity']['online'].get('parameters').get('serving_name')
            serving_url = [url for url in deployment_details['entity'].get('status').get('serving_urls') if serving_name == url.split('/')[-2]][0]

            if serving_url:
                return serving_url
            else:
                raise MissingValue(
                    u'Getting serving url for deployment failed. This functionality is available only for sync deployments with serving name.')

        except Exception as e:
            raise WMLClientError(
                u'Getting serving url for deployment failed. This functionality is available only for sync deployments with serving name.',
                e)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, deployment_uid):
        """
            Delete deployment.

            **Parameters**

            .. important::
                #. **deployment uid**:  Unique Id of  Deployment\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.deployments.delete(deployment_uid)

        """
        ##To be removed once deployments adds support for deployments
        self._client._check_if_space_is_set()

        deployment_uid = str_type_conv(deployment_uid)
        Deployments._validate_type(deployment_uid, u'deployment_uid', STR_TYPE, True)

        if deployment_uid is not None and not is_uid(deployment_uid):
            raise WMLClientError(u'\'deployment_uid\' is not an uid: \'{}\''.format(deployment_uid))

        deployment_url = self._client.service_instance._href_definitions.get_deployment_href(deployment_uid)

        if not self._ICP:
            response_delete = requests.delete(
                deployment_url,
                params=self._client._params(),
                headers=self._client._get_headers())
        else:
            if self._client.ICP_PLATFORM_SPACES:
                response_delete = requests.delete(
                    deployment_url,
                    params=self._client._params(),
                    headers=self._client._get_headers(),
                    verify=False)
            else:
                response_delete = requests.delete(
                    deployment_url,
                    #TODO:  The below line needs to uncommeted after /v4/deployment accespts query param
                   # params=self._client._params(),
                    headers=self._client._get_headers(),
                    verify=False)

        return self._handle_response(204, u'deployment deletion', response_delete, False)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def score(self,deployment_id, meta_props, transaction_id=None):
        """
                    Make scoring requests against deployed artifact.

                    **Parameters**

                    .. important::
                        #. **deployment_id**:  Unique Id of the deployment to be scored\n
                           **type**: str\n
                        #. **meta_props**:  Meta props for scoring\n
                                            >>> Use client.deployments.ScoringMetaNames.show() to view the list of ScoringMetaNames.
                           **type**: dict\n
                        #. **transaction_id**:  transaction id to be passed with records during payload logging (optional)\n
                           **type**: str\n

                    **Output**

                    .. important::
                        **returns**: scoring result containing prediction and probability\n
                        **return type**: dict\n

                    .. note::

                            * client.deployments.ScoringMetaNames.INPUT_DATA is the only metaname valid for sync scoring.\n
                            * The valid payloads for scoring input are either list of values, pandas or numpy dataframes.

                    **Example**

                     >>> scoring_payload = {wml_client.deployments.ScoringMetaNames.INPUT_DATA:
                     >>>        [{'fields':
                     >>>            ['GENDER','AGE','MARITAL_STATUS','PROFESSION'],
                     >>>            'values': [
                     >>>                ['M',23,'Single','Student'],
                     >>>                ['M',55,'Single','Executive']
                     >>>                ]
                     >>>         }
                     >>>       ]}
                     >>> predictions = client.deployments.score(deployment_id, scoring_payload)
                     >>> predictions = client.deployments.score(deployment_id, scoring_payload,async=True)

        """
        ##To be removed once deployments adds support for deployments
        self._client._check_if_space_is_set()

        Deployments._validate_type(deployment_id, u'deployment_id', STR_TYPE, True)
        Deployments._validate_type(meta_props, u'meta_props', dict, True)

        if meta_props.get(self.ScoringMetaNames.INPUT_DATA) is None:
            raise WMLClientError("Scoring data input 'ScoringMetaNames.INPUT_DATA' is mandatory for synchronous "
                                 "scoring")

        scoring_data = meta_props[self.ScoringMetaNames.INPUT_DATA]

        if scoring_data is not None:
            score_payload = []
            for each_score_request in scoring_data:
                lib_checker.check_lib(lib_name='pandas')
                import pandas as pd

                scoring_values = each_score_request["values"]
                # Check feature types, currently supporting pandas df, numpy.ndarray, python lists and Dmatrix
                if (isinstance(scoring_values, pd.DataFrame)):
                    fields_names = scoring_values.columns.values.tolist()
                    values = scoring_values.values.tolist()

                    each_score_request["values"] = values
                    if fields_names is not None:
                        each_score_request["fields"] = fields_names


                ## If payload is a numpy dataframe

                elif (isinstance(scoring_values, np.ndarray)):

                    values = scoring_values.tolist()
                    each_score_request["values"] = values


                # else:
                #     payload_score["values"] = each_score_request["values"]
                #     if "fields" in each_score_request:
                #         payload_score["fields"] = each_score_request["fields"]
                score_payload.append(each_score_request)

            ##See if it is scoring or DecisionOptimizationJob

        payload = {}

        payload["input_data"] = score_payload

        headers = self._client._get_headers()

        if transaction_id is not None:
            headers.update({'x-global-transaction-id': transaction_id})

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            scoring_url = self._wml_credentials["url"] + "/ml/v4/deployments/" + deployment_id + "/predictions"
        else:
            scoring_url = self._wml_credentials["url"] + "/v4/deployments/" + deployment_id + "/predictions"

        if not self._ICP:
            if self._client.CLOUD_PLATFORM_SPACES:
                params = self._client._params()
                del params[u'space_id']
                response_scoring = self.session.post(
                    scoring_url,
                    json=payload,
                    params=params,  # version parameter is mandatory
                    headers=headers)
            else:
                response_scoring = self.session.post(
                    scoring_url,
                    json=payload,
                    headers=headers)
        else:
            if self._client.ICP_PLATFORM_SPACES:
                params = self._client._params()
                del params[u'space_id']
                response_scoring = self.session.post(
                    scoring_url,
                    json=payload,
                    params=params,  # version parameter is mandatory
                    headers=headers,
                    verify=False)
            else:
                response_scoring = self.session.post(
                    scoring_url,
                    json=payload,
                    headers=headers,
                    verify=False)

        return self._handle_response(200, u'scoring', response_scoring)

        #########################################

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_download_url(deployment_details):
        """
        Get deployment_download_url from deployment details.

        **Parameters**

        .. important::

            #. **deployment_details**: Created deployment details.\n
               **type**: dict\n

        **Output**

        .. important::
            **returns**: Deployment download URL that is used to get file deployment (for example: Core ML).\n
            **return type**: str\n

        **Example**

            >>> deployment_url = client.deployments.get_download_url(deployment)
        """
        Deployments._validate_type(deployment_details, u'deployment_details', dict, True)
        try:
            virtual_deployment_detaails = deployment_details.get(u'entity').get(u'status').get(u'virtual_deployment_downloads')
            if virtual_deployment_detaails is not None:
                url = virtual_deployment_detaails[0].get(u'url')
            else:
                url = None
        except Exception as e:
            raise WMLClientError(u'Getting download url from deployment details failed.', e)

        if url is None:
            raise MissingValue(u'deployment_details.entity.virtual_deployment_downloads.url')

        return url

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, virtual_deployment_uid, filename=None):
        """
        Downloads file deployment of specified deployment Id. Currently supported format is Core ML.

        **Parameters**

        .. important::

            #. **virtual_deployment_uid**:  Unique Id of virtual deployment.\n
               **type**: str\n
            #. **filename**: filename of downloaded archive (optional).\n
               **type**: str\n

        **Output**

        .. important::
            **returns**: Path to downloaded file.\n
            **return type**: str\n

        **Example**

            >>> client.deployments.download(virtual_deployment_uid)
        """
        self._client._check_if_space_is_set()
        virtual_deployment_uid = str_type_conv(virtual_deployment_uid)
        Deployments._validate_type(virtual_deployment_uid, u'deployment_uid', STR_TYPE, False)

        if virtual_deployment_uid is not None and not is_uid(virtual_deployment_uid):
            raise WMLClientError(u'\'deployment_uid\' is not an uid: \'{}\''.format(virtual_deployment_uid))
        params = self._client._params()
        details = self.get_details(virtual_deployment_uid)
        download_url = self.get_download_url(details)
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_35 or self._client.ICP_40:
            download_url = download_url + "&version=" + self._client.version_param
            params = None
        if not self._ICP:
            # if self._client.CLOUD_PLATFORM_SPACES:
            #     download_url = download_url + "&version=" + self._client.version_param
            #     response = requests.get(
            #         download_url,
            #         headers=self._client._get_headers()
            #     )
            # else:
            response = requests.get(
                    download_url,
                    params=params,
                    headers=self._client._get_headers()
                )
        else:
            response = requests.get(
                download_url,
                params=params,
                headers=self._client._get_headers(),
                verify=False
            )
        if filename is None:
            filename = 'mlartifact.tar.gz'

        if response.status_code == 200:
            with open(filename, "wb") as new_file:
                new_file.write(response.content)
                new_file.close()

                print_text_header_h2(
                    u'Successfully downloaded deployment file: ' + str(filename))

                return filename
        else:
            raise WMLClientError(u'Unable to download deployment content: ' + response.text)

    # @docstring_parameter({'str_type': STR_TYPE_NAME})
    # def score(self, deployment_id, meta_props, transaction_id=None):
    #         """
    #             Make scoring requests against deployed artifact.
    #
    #             **Parameters**
    #
    #             .. important::
    #                 #. **deployment_id**:  UID of the deployment to be scored\n
    #                    **type**: str\n
    #                 #. **meta_props**:  Meta props for scoring\n
    #                                     >>> Use client.deployments.ScoringMetaNames.show() to view the list of ScoringMetaNames.
    #                    **type**: dict\n
    #                 #. **transaction_id**:  transaction id to be passed with records during payload logging (optional)\n
    #                    **type**: str\n
    #
    #             **Output**
    #
    #             .. important::
    #                 **returns**: scoring result containing prediction and probability\n
    #                 **return type**: dict\n
    #
    #             .. note::
    #
    #                     * client.deployments.ScoringMetaNames.SCORING_INPUT_DATA is the only metaname valid for sync scoring.\n
    #
    #             **Example**
    #
    #              >>> scoring_payload = {wml_client.deployments.ScoringMetaNames.INPUT_DATA: [{'fields': ['GENDER','AGE','MARITAL_STATUS','PROFESSION'], 'values': [['M',23,'Single','Student'],['M',55,'Single','Executive']]}]}
    #              >>> predictions = client.deployments.score(deployment_id, scoring_payload)
    #              >>> predictions = client.deployments.score(deployment_id, scoring_payload,async=True)
    #         """
    #         Deployments._validate_type(deployment_id, u'scoring_url', STR_TYPE, True)
    #         Deployments._validate_type(meta_props, u'meta_props', dict, True)
    #
    #
    #         score_payload = {}
    #         if meta_props[self.ScoringMetaNames.INPUT_DATA] is None:
    #             raise WMLClientError("Scoring data input is mandatory for synchronous scoring")
    #
    #         score_payload["input_data"] = meta_props[self.ScoringMetaNames.INPUT_DATA]
    #         payload = score_payload
    #         Deployments._validate_type(deployment_id, u'scoring_url', STR_TYPE, True)
    #         Deployments._validate_type(payload, u'payload', dict, True)
    #
    #         headers = self._client._get_headers()
    #
    #         if transaction_id is not None:
    #             headers.update({'x-global-transaction-id': transaction_id})
    #         # making change - connection keep alive
    #         scoring_url = self._wml_credentials["url"] + "/v4/deployments/"+deployment_id+"/predictions"
    #         if not self._ICP:
    #             response_scoring = self.session.post(
    #                 scoring_url,
    #                 json=payload,
    #                 headers=headers)
    #         else:
    #             response_scoring = self.session.post(
    #                 scoring_url,
    #                 json=payload,
    #                 headers=headers,
    #                 verify=False)
    #
    #         return self._handle_response(200, u'scoring', response_scoring)

    def list(self, limit=None):
        """
           List deployments. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all deployments in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.deployments.list()

        """
        ##To be removed once deployments adds support for deployments
        self._client._check_if_space_is_set()

        details = self.get_details(limit=limit)
        resources = details[u'resources']

        values = []
        index = 0

        for m in resources:
            # Deployment service currently doesn't support limit querying
            # As a workaround, its filtered in python client
            # Ideally this needs to be on the server side
            if limit is not None and index == limit:
                break

            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                if 'guid' in m[u'metadata']:
                    values.append((m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'entity'][u'status']['state'],
                                  m[u'metadata'][u'created_at'], self._get_deployable_asset_type(m)))
                else:
                    values.append((m[u'metadata'][u'id'], m[u'entity'][u'name'], m[u'entity'][u'status']['state'],
                                  m[u'metadata'][u'created_at'], self._get_deployable_asset_type(m)))
            else:
                if 'guid' in m[u'metadata']:
                    values.append((m[u'metadata'][u'guid'], m[u'entity'][u'name'], m[u'entity'][u'status']['state'],
                                  m[u'metadata'][u'created_at']))
                else:
                    values.append((m[u'metadata'][u'id'], m[u'entity'][u'name'], m[u'entity'][u'status']['state'],
                                  m[u'metadata'][u'created_at']))

            index = index + 1

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            self._list(values, [u'GUID', u'NAME', u'STATE', u'CREATED', u'ARTIFACT_TYPE'], limit, 50)
        else:
            self._list(values, [u'GUID', u'NAME', u'STATE', u'CREATED'], limit, 50)

    def list_jobs(self,limit=None):
        """
        List the async deployment jobs. If limit is set to None there will be only first 50 records shown.

        **Parameters**

        .. important::
            #. **limit**:  limit number of fetched records\n
               **type**: int\n

        **Output**

        .. important::

            This method only prints the list of all async jobs in a table format.\n
                **return type**: None\n

        .. note::
            * This method list only async deployment jobs created for WML deployment.

        **Example**

         >>> client.deployments.list_jobs()
        """

        details = self.get_job_details(limit=limit)
        resources = details[u'resources']
        values = []
        index = 0

        for m in resources:
            # Deployment service currently doesn't support limit querying
            # As a workaround, its filtered in python client
            if limit is not None and index == limit:
                break

            if 'scoring' in m['entity']:
                state = m['entity']['scoring']['status']['state']
            else:
                state = m['entity']['decision_optimization']['status']['state']

            if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                deploy_id = m['entity']['deployment']['id']
                values.append((m[u'metadata'][u'id'], state, m[u'metadata'][u'created_at'], deploy_id))
            else:
                deploy_id = m['entity']['deployment']['href'].split("/")[3]
                values.append((m[u'metadata'][u'guid'], state, m[u'metadata'][u'created_at'], deploy_id))

            index = index + 1

        self._list(values, [u'JOB-UID', u'STATE', u'CREATED',u'DEPLOYMENT-ID'], limit, 50)

    def _get_deployable_asset_type(self, details):
        url = details[u'entity'][u'asset']['id']
        if 'model' in url:
            return 'model'
        elif 'function' in url:
            return 'function'
        else:
            return 'unknown'

    def update(self, deployment_uid, changes):
        """
                Updates existing deployment metadata. If ASSET is patched, then 'id' field is mandatory and
                it starts a deployment with the provided asset id/rev. Deployment id remains same

                **Parameters**

                .. important::

                    #. **deployment_uid**: Unique Id of deployment which  should be updated\n
                       **type**: str\n
                    #. **changes**:  elements which should be changed, where keys are ConfigurationMetaNames\n
                       **type**: dict\n

                **Output**

                .. important::

                    **returns**: metadata of updated deployment\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>> client.deployments.ConfigurationMetaNames.NAME:"updated_Deployment",
                 >>> client.deployments.ConfigurationMetaNames.ASSET: { "id": "ca0cd864-4582-4732-b365-3165598dc945", "rev":"2" }
                 >>> }
                 >>>
                 >>> deployment_details = client.deployments.update(deployment_uid, changes=metadata)

        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        ##To be removed once deployments adds support for deployments
        self._client._check_if_space_is_set()

        ret202 = False

        ## In case of passing 'AUTO_ONLINE_DEPLOYMENT' as true, we need to poll for deployment to be either 'deploy_success' or 'update_success'.

        deployment_uid = str_type_conv(deployment_uid)
        Deployments._validate_type(deployment_uid, 'deployment_uid', STR_TYPE, True)

        if 'asset' in changes and not changes[u'asset']:
            msg = "ASSETS cannot be empty. 'id' and 'rev' fields are supported. 'id' is mandatory"
            print(msg)
            raise WMLClientError(msg)

        # if changes.get('asset') is not None and (changes.get('name') is not None or changes.get('description') is not None):
        patch_job = changes.get('asset') is not None or self.ConfigurationMetaNames.SERVING_NAME in changes or self.ConfigurationMetaNames.OWNER in changes

        patch_job_field = None
        if patch_job:
            if changes.get('asset') is not None:
                patch_job_field = "ASSET"
            elif self.ConfigurationMetaNames.SERVING_NAME in changes:
                patch_job_field = "SERVING_NAME"
            elif self.ConfigurationMetaNames.OWNER in changes:
                patch_job_field = "OWNER"

            if patch_job_field is None:
                raise WMLClientError("Unexpected patch job element.")

        if patch_job and (len(changes) > 1):
            msg = f"When {patch_job_field} is being updated/patched, other fields cannot be updated. If other fields are to be " \
                  f"updated, try without {patch_job_field} update. {patch_job_field} update triggers deployment with the new asset retaining " \
                  "the same deployment_id"
            print(msg)
            raise WMLClientError(msg)

        deployment_details = self.get_details(deployment_uid)
        serving_name_change = False
        new_serving_name = None
        if self.ConfigurationMetaNames.SERVING_NAME in changes:
            new_serving_name = changes.pop(self.ConfigurationMetaNames.SERVING_NAME)
            serving_name_change = True

        patch_payload = self.ConfigurationMetaNames._generate_patch_payload(deployment_details, changes, with_validation=True)

        if serving_name_change:
            replace = 'serving_name' in deployment_details['entity'].get('online').get('parameters', [])
            patch_payload.append({
                "op": "replace" if replace else "add",
                "path": "/online/parameters",
                "value": {"serving_name": new_serving_name}})

        ## As auto_online_deployment and auto_redeploy values are passed as 'bool' but service needs them in 'str' format to patch.
        for ele in patch_payload:
            if('auto_online_deployment' in ele['path'] or 'auto_redeploy' in ele['path']):
                if(ele['value']==True):
                    ele['value']='true'
                else:
                    ele['value']='false'

        url = self._client.service_instance._href_definitions.get_deployment_href(deployment_uid)

        if not self._ICP:
            response = requests.patch(url, json=patch_payload, params=self._client._params(), headers=self._client._get_headers())
        else:
            response = requests.patch(url, json=patch_payload, params=self._client._params(), headers=self._client._get_headers(), verify=False)

        if patch_job and response.status_code == 202:
            updated_details = self._handle_response(202, u'deployment asset patch', response)

            ret202 = True

            print(f"Since {patch_job_field} is patched, deployment with new asset id/rev is being started. " \
                  "Monitor the status using deployments.get_details(deployment_uid) api")
        elif response.status_code == 202:
            updated_details = self._handle_response(202, u'deployment scaling', response)
            ret202 = True
        else:
            updated_details = self._handle_response(200, u'deployment patch', response)

        if('auto_online_deployment' in changes):
           if response is not None:
               if response.status_code == 200:
                   deployment_details = self.get_details(deployment_uid)

                   import time
                   print_text_header_h1(u' deployment update for uid: \'{}\' started'.format(deployment_uid))

                   status = deployment_details[u'entity'][u'status'][u'state']

                   with StatusLogger(status) as status_logger:
                       while True:
                           time.sleep(5)
                           deployment_details = self.get_details(deployment_uid)
                           status = deployment_details[u'entity'][u'status'][u'state']
                           status_logger.log_state(status)

                           if status != u'DEPLOY_IN_PROGRESS' and status != u'UPDATE_IN_PROGRESS':
                               break

                   if status == u'DEPLOY_SUCCESS' or status == u'UPDATE_SUCCESS':
                       print(u'')
                       print_text_header_h2(
                           u'Successfully finished deployment update, deployment_uid=\'{}\''.format(deployment_uid))
                       return deployment_details
                   else:
                       print_text_header_h2(u'Deployment update failed')
                       self._deployment_status_errors_handling(deployment_details, 'update', deployment_uid)
               else:
                   error_msg = u'Deployment update failed'
                   reason = response.text
                   print(reason)
                   print_text_header_h2(error_msg)
                   raise WMLClientError(error_msg + u'. Error: ' + str(response.status_code) + '. ' + reason)

        if not ret202:
            return updated_details

    ## Below functions are for async scoring. They are just dummy functions.
    def _score_async(self,deployment_uid, scoring_payload,
                     transaction_id=None,
                     retention=None):

        Deployments._validate_type(deployment_uid, u'deployment_id', STR_TYPE, True)
        Deployments._validate_type(scoring_payload, u'scoring_payload', dict, True)
        headers = self._client._get_headers()

        if transaction_id is not None:
            headers.update({'x-global-transaction-id': transaction_id})
        # making change - connection keep alive
        scoring_url = self._client.service_instance._href_definitions.get_async_deployment_job_href()
        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            params = self._client._params()
        else:
            params = None

        if not self._client.ICP and retention is not None:
            if not isinstance(retention, int) or retention < -1:
                raise TypeError("`retention` takes integer values greater or equal than -1.")
            params.update({'retention': retention})

        if not self._ICP:
            response_scoring = self.session.post(
                scoring_url,
                params=params,
                json=scoring_payload,
                headers=headers)
        else:
            response_scoring = self.session.post(
                scoring_url,
                params = params,
                json=scoring_payload,
                headers=headers,
                verify=False)

        return self._handle_response(202, u'scoring asynchronously', response_scoring)

    def create_job(self, deployment_id, meta_props, retention=None, transaction_id=None):
        """
            Create an asynchronous deployment job.

            **Parameters**

            .. important::

                #. **deployment_id**:  Unique Id of Deployment\n
                   **type**: str\n

                #. **meta_props**: metaprops. To see the available list of metanames use:\n
                                   >>> client.deployments.ScoringMetaNames.get() or client.deployments.DecisionOptimizationmetaNames.get()

                   **type**: dict\n

                #. **retention**: How many job days job meta should be retained (optional). Takes integer values >= -1.
                                  Supported only on Cloud.\n
                   **type**:: int\n

            **Output**

            .. important::

                **returns**: metadata of the created async deployment job\n
                **return type**: dict\n

            .. note::

                * The valid payloads for scoring input are either list of values, pandas or numpy dataframes.

            **Example**

             >>> scoring_payload = {wml_client.deployments.ScoringMetaNames.INPUT_DATA: [{'fields': ['GENDER','AGE','MARITAL_STATUS','PROFESSION'], 'values': [['M',23,'Single','Student'],['M',55,'Single','Executive']]}]}
             >>> async_job = client.deployments.create_job(deployment_id, scoring_payload)

         """

        WMLResource._chk_and_block_create_update_for_python36(self)
        Deployments._validate_type(deployment_id, u'deployment_uid', STR_TYPE, True)
        Deployments._validate_type(meta_props, u'meta_props', dict, True)

        scoring_data = None
        flag = 0 ## To see if it is async scoring or DecisionOptimization Job
        deployment_details = self.get_details(deployment_id)

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            asset = deployment_details["entity"]["asset"]['id']
        else:
            asset = deployment_details["entity"]["asset"]["href"].split("/")[-1]

        do_model = False
        if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            asset_details = self._client.data_assets.get_details(asset)
            if 'wml_model' in asset_details["entity"] and 'type' in asset_details['entity']['wml_model']:
                if 'do' in asset_details['entity']['wml_model']['type']:
                    do_model = True
        else:
            asset_details = self._client.repository.get_details(asset)
            if "type" in asset_details["entity"]:
                if "do" in asset_details["entity"]["type"]:
                    do_model = True

        if do_model:
            payload = self.DecisionOptimizationMetaNames._generate_resource_metadata(meta_props,
                                                                                     with_validation=True,
                                                                                     client=self._client)
            flag = 1
        else:
            payload = self.ScoringMetaNames._generate_resource_metadata(meta_props, with_validation=True,
                                                                        client=self._client)

        if "scoring" in payload and "input_data" in payload["scoring"]:
                scoring_data = payload["scoring"]["input_data"]

        if "decision_optimization" in payload and "input_data" in payload["decision_optimization"]:
                scoring_data = payload["decision_optimization"]["input_data"]

        if scoring_data is not None:
            score_payload = []
            for each_score_request in scoring_data:
                lib_checker.check_lib(lib_name='pandas')
                import pandas as pd
                if "values" in each_score_request:
                    scoring_values = each_score_request["values"]
                    # Check feature types, currently supporting pandas df, numpy.ndarray, python lists and Dmatrix
                    if (isinstance(scoring_values, pd.DataFrame)):
                        fields_names = scoring_values.columns.values.tolist()
                        values = scoring_values.where(pd.notnull(scoring_values), None).values.tolist() #replace nan with None

                        each_score_request["values"] = values
                        if fields_names is not None:
                            each_score_request["fields"] = fields_names


                    ## If payload is a numpy dataframe

                    elif (isinstance(scoring_values, np.ndarray)):

                        values = np.where(pd.notnull(scoring_values), scoring_values, None).tolist() #replace nan with None
                        each_score_request["values"] = values


                # else:
                #     payload_score["values"] = each_score_request["values"]
                #     if "fields" in each_score_request:
                #         payload_score["fields"] = each_score_request["fields"]
                score_payload.append(each_score_request)

            ##See if it is scoring or DecisionOptimizationJob

            if flag == 0:
                payload["scoring"]["input_data"] = score_payload
            if flag == 1:
                payload["decision_optimization"]["input_data"] = score_payload

        if self._client.ICP_30 or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            import copy
            if 'input_data_references' in meta_props:
                Deployments._validate_type(meta_props.get('input_data_references'), u'input_data_references', list,
                                           True)
                modified_input_data_references=False
                input_data = copy.deepcopy(meta_props.get('input_data_references'))
                for i, input_data_fields in enumerate(input_data):
                    if 'connection' not in input_data_fields:
                        modified_input_data_references=True
                        input_data_fields.update({'connection': {}})
                if modified_input_data_references:
                    if 'scoring' in payload:
                        payload['scoring'].update({'input_data_references': input_data})
                    else:
                        payload['decision_optimization'].update({'input_data_references': input_data})

            if 'output_data_reference' in meta_props:
                Deployments._validate_type(meta_props.get('output_data_reference'), u'output_data_reference', dict,
                                           True)

                output_data = copy.deepcopy(meta_props.get('output_data_reference'))
                if 'connection' not in output_data:  # and output_data.get('connection', None) is not None:
                    output_data.update({'connection': {}})
                    payload['scoring'].update({'output_data_reference': output_data})

            if 'output_data_references' in meta_props:
                Deployments._validate_type(meta_props.get('output_data_references'), u'output_data_references', list, True)
                output_data = copy.deepcopy(meta_props.get('output_data_references'))
                modified_output_data_references = False
                for i, output_data_fields in enumerate(output_data):
                    if 'connection' not in output_data_fields:
                        modified_output_data_references = True
                        output_data_fields.update({'connection': {}})
                if modified_output_data_references and 'decision_optimization' in payload:
                    payload['decision_optimization'].update({'output_data_references': output_data})

            payload.update({"deployment": {"id": deployment_id}})
            if 'hardware_spec' in meta_props:
                payload.update({"hardware_spec": meta_props[self.ConfigurationMetaNames.HARDWARE_SPEC]})
            if 'hybrid_pipeline_hardware_specs' in meta_props:
                payload.update({"hybrid_pipeline_hardware_specs": meta_props[self.ConfigurationMetaNames.HYBRID_PIPELINE_HARDWARE_SPECS]})

            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                payload.update({'space_id': self._client.default_space_id})

                if 'name' not in payload:
                    import uuid
                    payload.update({'name': 'name_' + str(uuid.uuid4())})
        else:
            payload.update({"deployment": {"href": "/v4/deployments/" + deployment_id}})

        return self._score_async(deployment_id, payload,
                                 transaction_id=transaction_id,
                                 retention=retention)

    def get_job_details(self, job_uid=None, limit=None):
        """
           Get information about your deployment job(s). If  deployment job_uid is not passed, all deployment jobs details are fetched.

           **Parameters**

           .. important::

                #. **job_uid**: Unique Job ID (optional)\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n

           **Output**

           .. important::
                **returns**: metadata of deployment job(s)\n
                **return type**: dict\n
                dict (if job_uid is not None) or {"resources": [dict]} (if job_uid is None)\n

           .. note::
                If job_uid is not specified, all deployment jobs metadata associated with the deployment Id is fetched\n


           **Example**

            >>> deployment_details = client.deployments.get_job_details()
            >>> deployments_details = client.deployments.get_job_details(job_uid=job_uid)

        """
        if job_uid is not None:
            Deployments._validate_type(job_uid, u'job_uid', STR_TYPE, True)
        url = self._client.service_instance._href_definitions.get_async_deployment_job_href()

        if self._ICP:
            if job_uid is not None:
                url = url + '/' + job_uid
                response = requests.get(
                    url,
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    verify=False
                )
                return self._handle_response(200, "async deployment job", response)
            else:
                return self._get_artifact_details(url, job_uid, limit, 'async deployment jobs')
        else:
            return self._get_artifact_details(url, job_uid, limit, 'async deployment jobs')

    def get_job_status(self, job_id):
        """
           Get the status of the deployment job.

           **Parameters**

           .. important::
                #. **job_id**: Unique Id of the deployment job\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: status of the deployment job\n
                **return type**: dict\n

           **Example**

            >>> job_status = client.deployments.get_job_status(job_uid)

        """

        job_details = self.get_job_details(job_id)

        if 'scoring' not in job_details['entity']:
            return job_details['entity']['decision_optimization']['status']
        return job_details['entity']['scoring']['status']

    def get_job_uid(self, job_details):
        """
           Get the Unique Id of the deployment job.

           **Parameters**

           .. important::
                #. **job_details**: metadata of the deployment job\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: Unique Id of the deployment job\n
                **return type**: str\n

           **Example**

            >>> job_details = client.deployments.get_job_details(job_uid=job_uid)
            >>> job_status = client.deployments.get_job_uid(job_details)

        """

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            return job_details['metadata']['id']
        else:
            return job_details['metadata']['guid']

    def get_job_href(self, job_details):
        """
           Get the href of the deployment job.

           **Parameters**

           .. important::
                #. **job_details**: metadata of the deployment job\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of the deployment job\n
                **return type**: str\n

           **Example**

            >>> job_details = client.deployments.get_job_details(job_uid=job_uid)
            >>> job_status = client.deployments.get_job_href(job_details)

        """

        if self._client.ICP_30 and not self._client.ICP_PLATFORM_SPACES:
            job_href = "/v4/deployment_jobs/" + job_details['metadata']['id']
            return job_href
        else:
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                return "/ml/v4/deployment_jobs/{}".format(job_details[u'metadata'][u'id'])
            else:
                return job_details['metadata']['href']


    def delete_job(self, job_uid, hard_delete=False):
        """
        Cancels a deployment job that is currenlty running.  This method is also be used to delete metadata
        details of the completed or canceled jobs when hard_delete parameter is set to True.

        **Parameters**

        .. important::

            #. **job_uid**:  Unique Id of deployment job which  should be canceled\n
               **type**: str\n

            #. **hard_delete**: specify True or False.
                                True - To delete the completed or canceled job.
                                False - To cancel the currently running deployment job. Default value is False.
               **type**: Boolean\n


        **Output**

        .. important::

            **returns**: status ("SUCCESS" or "FAILED")\n
            **return type**: str\n

        **Example**

         >>> client.deployments.delete_job(job_uid)

        """

        job_uid = str_type_conv(job_uid)

        Deployments._validate_type(job_uid, u'job_uid', STR_TYPE, True)

        if job_uid is not None and not is_uid(job_uid):
            raise WMLClientError(u'\'job_uid\' is not an uid: \'{}\''.format(job_uid))

        url = self._client.service_instance._href_definitions.get_async_deployment_jobs_href(job_uid)

        if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
            params = self._client._params()
        else:
            params = {}

        if hard_delete is True:
            params.update({'hard_delete': u'true'})

        if not self._ICP:
            response_delete = requests.delete(
                url,
                headers=self._client._get_headers(),
                params=params)
        else:
            response_delete = requests.delete(
                url,
                headers=self._client._get_headers(),
                params=params,
                verify=False)

        return self._handle_response(204, u'deployment async job deletion', response_delete, False)


    # def create_environment(self, meta_props):
    #     """
    #             Cancels a job.
    #
    #             **Parameters**
    #
    #             .. important::
    #
    #                 #. **job_uid**:  UID of job which  should be canceled\n
    #                    **type**: str\n
    #
    #             **Output**
    #
    #             .. important::
    #
    #                 **returns**: status ("SUCCESS" or "FAILED")\n
    #                 **return type**: str\n
    #
    #             **Example**
    #
    #              >>> client.deployments.delete_job(job_uid)
    #     """
    #
    #     Deployments._validate_type(meta_props, u'metaprops', dict, True)
    #
    #
    #     url = self._client.service_instance._href_definitions.get_envs_href()
    #     metadata = self.EnvironmentMetanames._generate_resource_metadata(meta_props)
    #
    #     if not self._ICP:
    #         response_post = requests.post(url, json=metadata,
    #                                           headers=self._client._get_headers())
    #
    #
    #     else:
    #         response_post = requests.post(url, json=metadata,
    #                                       headers=self._client._get_headers(), verify=False)
    #
    #     return self._handle_response(201, u'deployment job deletion', response_post, False)
    #
    # def delete_environment(self, env_id):
    #     """
    #             Cancels a job.
    #
    #             **Parameters**
    #
    #             .. important::
    #
    #                 #. **job_uid**:  UID of job which  should be canceled\n
    #                    **type**: str\n
    #
    #             **Output**
    #
    #             .. important::
    #
    #                 **returns**: status ("SUCCESS" or "FAILED")\n
    #                 **return type**: str\n
    #
    #             **Example**
    #
    #              >>> client.deployments.delete_job(job_uid)
    #     """
    #
    #     env_id = str_type_conv(env_id)
    #
    #     Deployments._validate_type(env_id, u'env_uid', STR_TYPE, True)
    #
    #
    #     if env_id is not None and not is_uid(env_id):
    #         raise WMLClientError(u'\'env_id\' is not an uid: \'{}\''.format(env_id))
    #
    #
    #     url = self._client.service_instance._href_definitions.get_env_href(env_id)
    #
    #     if not self._ICP:
    #         response_delete = requests.delete(
    #             url,
    #             headers=self._client._get_headers())
    #     else:
    #         response_delete = requests.delete(
    #             url,
    #             headers=self._client._get_headers(),
    #             verify=False)
    #
    #     return self._handle_response(204, u'environment deletion', response_delete, False)
    #
    # def get_environment_uid(self, env_details):
    #     """
    #        Get the UID of the environment.
    #
    #        **Parameters**
    #
    #        .. important::
    #             #. **job_details**: metadata of the job\n
    #                **type**: dict\n
    #
    #        **Output**
    #
    #        .. important::
    #             **returns**: UID of the job\n
    #             **return type**: str\n
    #
    #        **Example**
    #
    #         >>> job_details = client.deployments.get_job_details(deployment_uid=deployment_uid,job_uid=job_uid)
    #         >>> job_status = client.deployments.get_job_uid(job_details)
    #     """
    #
    #     return env_details['metadata']['guid']
    #
    # def get_environment_href(self, env_details):
    #     """
    #        Get the href of the job.
    #
    #        **Parameters**
    #
    #        .. important::
    #             #. **job_details**: metadata of the job\n
    #                **type**: dict\n
    #
    #        **Output**
    #
    #        .. important::
    #             **returns**: href of the job\n
    #             **return type**: str\n
    #
    #        **Example**
    #
    #         >>> job_details = client.deployments.get_job_details(deployment_uid=deployment_uid,job_uid=job_uid)
    #         >>> job_status = client.deployments.get_job_href(job_details)
    #     """
    #
    #     return env_details['metadata']['href']
    #
    # def get_environment_details(self,env_id=None,limit=None):
    #     """
    #        Get information about your job(s). If job_uid is not passed, all jobs details are fetched.
    #
    #        **Parameters**
    #
    #        .. important::
    #
    #             #. **job_uid**: Job UID (optional)\n
    #                **type**: str\n
    #             #. **limit**:  limit number of fetched records (optional)\n
    #                **type**: int\n
    #
    #        **Output**
    #
    #        .. important::
    #             **returns**: metadata of job(s)\n
    #             **return type**: dict\n
    #             dict (if job UID is not None) or {"resources": [dict]} (if job UID is None)\n
    #
    #        .. note::
    #             If job UID is not specified, all jobs metadata associated with the deployment UID is fetched\n
    #
    #
    #        **Example**
    #
    #         >>> deployment_details = client.deployments.get_job_details()
    #         >>> deployments_details = client.deployments.get_job_details(job_uid=job_uid)
    #     """
    #     #Deployments._validate_type(job_uid, u'deployment_uid', STR_TYPE, True)
    #     #Deployments._validate_type(job_uid, u'job_uid', STR_TYPE, True)
    #     url = self._client.service_instance._href_definitions.get_envs_href()
    #     return self._get_artifact_details(url, env_id, limit, 'get environments details')
    #
    # def list_environments(self,limit=None):
    #     """
    #        List the async jobs. If limit is set to None there will be only first 50 records shown.
    #
    #        **Parameters**
    #
    #        .. important::
    #             #. **limit**:  limit number of fetched records\n
    #                **type**: int\n
    #
    #        **Output**
    #
    #        .. important::
    #             This method only prints the list of all async jobs in a table format.\n
    #             **return type**: None\n
    #
    #        **Example**
    #
    #         >>> client.deployments.list_jobs()
    #     """
    #     details = self.get_environment_details()
    #     resources = details[u'resources']
    #     values = [(m[u'metadata'][u'guid'], m[u'entity'][u'name'],
    #                m[u'metadata'][u'created_at']) for m in resources]
    #
    #     self._list(values, [u'GUID', u'NAME', u'CREATED'], None, 50)




"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
from abc import ABC, abstractmethod


class Serializer(ABC):
    """
    Abstract class for Serializer
    """
    @abstractmethod
    def serialize(self, msg):
        """
        Serialize a message

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialized byte stream
        :rtype: `b[]`
        """
        pass

    @abstractmethod
    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        pass



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tarfile, gzip, shutil


class CompressionUtil(object):
    @staticmethod
    def create_tar(path, arcname, archive_path):
        tar = tarfile.open(archive_path, "w")
        tar.add(path, arcname)
        tar.close()

    @staticmethod
    def extract_tar(archive_path, path):
        tar = tarfile.open(archive_path)
        tar.extractall(path)
        tar.close()

    @staticmethod
    def compress_file_gzip(filepath, gzip_filepath):
        with open(filepath, 'rb') as f_in, gzip.open(gzip_filepath, 'wb+') as f_out:
            shutil.copyfileobj(f_in, f_out)

    @staticmethod
    def decompress_file_gzip(gzip_filepath, filepath):
        with gzip.open(gzip_filepath, 'rb') as f:
            content = f.read()
            output_f = open(filepath, 'wb+')
            output_f.write(content)
            output_f.close()



import unittest

import ibm_boto3
from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.helpers.connections import DataConnection, S3Location
from ibm_watson_machine_learning.utils.autoai.errors import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket, is_cp4d
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import AbstractTestAutoAIAsync, \
    AbstractTestWebservice

from ibm_watson_machine_learning.tests.utils import get_wml_credentials, is_cp4d

from ibm_watson_machine_learning.utils.autoai.enums import PredictionType, Metrics, ClassificationAlgorithms


@unittest.skipIf(not is_cp4d(), "Not supported on cloud")
class TestAutoAIRemote(AbstractTestAutoAIAsync, unittest.TestCase):
    """
    The test can be run on CPD only
    """

    cos_resource = None
    data_location = './autoai/data/breast_cancer.csv'

    data_cos_path = 'data/breast_cancer.csv'

    SPACE_ONLY = False

    OPTIMIZER_NAME = "breast_cancer test sdk"

    target_space_id = None

    experiment_info = dict(
        name=OPTIMIZER_NAME,
        prediction_type=PredictionType.BINARY,
        prediction_column='diagnosis',
        positive_label='M',
        scoring=Metrics.AVERAGE_PRECISION_SCORE,
        max_number_of_estimators=1,
        include_only_estimators=[ClassificationAlgorithms.DT]

    )

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        cls.wml_credentials = get_wml_credentials()
        tmp_client = APIClient(wml_credentials=cls.wml_credentials.copy())
        try:
            del cls.wml_credentials['username']
        except:
            pass
        try:
            del cls.wml_credentials['password']
        except:
            pass
        try:
            del cls.wml_credentials['apikey']
        except:
            pass
        try:
            del cls.wml_credentials['bedrock_url']
        except:
            pass
        cls.wml_credentials['token'] = tmp_client.wml_token
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

        cls.project_id = cls.wml_credentials.get('project_id')

    def test_00d_prepare_data_asset(self):
        asset_details = self.wml_client.data_assets.create(
            name=self.data_location.split('/')[-1],
            file_path=self.data_location)

        TestAutoAIRemote.asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(self.asset_id, str)

    def test_02_data_reference_setup(self):
        TestAutoAIRemote.data_connection = DataConnection(data_asset_id=self.asset_id)
        TestAutoAIRemote.results_connection = None

        self.assertIsNotNone(obj=TestAutoAIRemote.data_connection)
        self.assertIsNone(obj=TestAutoAIRemote.results_connection)

    def test_99_delete_data_asset(self):
        self.wml_client.data_assets.delete(self.asset_id)

        with self.assertRaises(WMLClientError):
            self.wml_client.data_assets.get_details(self.asset_id)


if __name__ == '__main__':
    unittest.main()



#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.function_artifact_loader  import FunctionArtifactLoader


class FunctionLoader(FunctionArtifactLoader):
    """
        Returns  Generic function instance associated with this function artifact.

        :return: function
        :rtype:
        """

    def function_instance(self):
        """
         :return: returns function path
         """
        return self.load()

    def download_function(self,path):
        """
         :return: returns function path
         """
        return self.load(path)




# coding: utf-8

"""
Copyright 2016 SmartBear Software

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Ref: https://github.com/swagger-api/swagger-codegen
"""

#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems


class LibrariesDefinitionInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None,description=None, version=None, platform=None):
        """
        LibrariesDefinitionInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'description': 'str',
            'version': 'str',
            'platform': 'LibrariesDefinitionInputPlatform',
        }

        self.attribute_map = {
            'name': 'name',
            'description': 'description',
            'version': 'version',
            'platform': 'platform',
        }

        self._name = name
        self._description = description
        self._version = version
        self._platform = platform

    @property
    def name(self):
        """
        Gets the name of this LibrariesDefinitionInput.


        :return: The name of this LibrariesDefinitionInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this LibrariesDefinitionInput.


        :param name: The name of this LibrariesDefinitionInput.
        :type: str
        """
        self._name = name

    @property
    def description(self):
        """
        Gets the description of this LibrariesDefinitionInput.


        :return: The description of this LibrariesDefinitionInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this LibrariesDefinitionInput.


        :param description: The description of this LibrariesDefinitionInput.
        :type: str
        """
        self._description = description

    @property
    def version(self):
        """
        Gets the version of this LibrariesDefinitionInput.


        :return: The version of this LibrariesDefinitionInput.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """
        Sets the version of this LibrariesDefinitionInput.


        :param version: The version of this LibrariesDefinitionInput.
        :type: str
        """
        self._version = version

    @property
    def platform(self):
        """
        Gets the platform of this LibrariesDefinitionInput.


        :return: The platform of this LibrariesDefinitionInput.
        :rtype: str
        """
        return self._platform

    @platform.setter
    def platform(self, platform):
        """
        Sets the platform of this LibrariesDefinitionInput.


        :param platform: The platform of this LibrariesDefinitionInput.
        :type: str
        """
        self._platform = platform

    @property
    def platform_versions(self):
        """
        Gets the platform_versions of this LibrariesDefinitionInput.


        :return: The platform_versions of this LibrariesDefinitionInput.
        :rtype: list[str]
        """
        return self._platform_versions

    @platform_versions.setter
    def platform_versions(self, platform_versions):
        """
        Sets the platform_versions of this LibrariesDefinitionInput.


        :param platform_versions: The platform_versions of this LibrariesDefinitionInput.
        :type: list[str]
        """
        self._platform_versions = platform_versions

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
JSON based serialization
"""
import jsonpickle
from ibmfl.message.message import Message
from ibmfl.message.serializer import Serializer


class JSONSerializer(Serializer):
    """
    Class for JSON Serializer
    """

    def serialize(self, msg):
        """
        Serialize a message using JSON

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialize byte stream
        :rtype: `b[]`
        """
        msg_header = msg.get_header()
        data = msg.get_data()
        json_str_obj = jsonpickle.encode(
            {
                'header': msg_header,
                'data': data,
            })
        return json_str_obj.encode()  # return a byte stream

    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        json_str_obj = serialized_byte_stream.decode()
        data_dict = jsonpickle.decode(json_str_obj)

        msg = Message(data=data_dict['data'])
        msg.set_header(data_dict['header'])

        return msg



import unittest

import logging
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *

class TestModelDefn(unittest.TestCase):
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestModelDefn.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.project_id = get_project_id()

        self.client.set.default_project(self.project_id)

        # print(self.client.service_instance.get_details())

        print("project_id: ", self.project_id)

    def test_01_create_model_defn_asset(self):

        self.client.model_definitions.ConfigurationMetaNames.show()
        sw_spec_uid = self.client.software_specifications.get_uid_by_name("ai-function_0.1-py3.6")

        print("sw_spec_uid: ", sw_spec_uid)

        meta_props = {
            self.client.model_definitions.ConfigurationMetaNames.NAME: "Test Model Definition",
            self.client.model_definitions.ConfigurationMetaNames.VERSION: "1.0",
            self.client.model_definitions.ConfigurationMetaNames.PLATFORM: {"name": "python", "versions": ["3.5"]},
            self.client.model_definitions.ConfigurationMetaNames.COMMAND: "test_command"
        }

        model_defn_details = self.client.model_definitions.store(meta_props = meta_props,
                                                                 model_definition="artifacts/pytorch_onnx_v_1.1.zip")

        print(model_defn_details)
        TestModelDefn.model_defn_asset_id = self.client.model_definitions.get_id(model_defn_details)
        TestModelDefn.model_defn_asset_url = self.client.model_definitions.get_href(model_defn_details)
        TestModelDefn.logger.info("model_defn asset ID:" + str(TestModelDefn.model_defn_asset_id))
        TestModelDefn.logger.info("model_defn asset URL:" + str(TestModelDefn.model_defn_asset_url))
        self.assertIsNotNone(TestModelDefn.model_defn_asset_id)
        self.assertIsNotNone(TestModelDefn.model_defn_asset_url)


    def test_02_download_model_defn_content(self):
        try:
            os.remove('test_model_defn_asset.zip')
        except:
            pass
        self.client.model_definitions.download(TestModelDefn.model_defn_asset_id, filename='test_model_defn_asset.zip')
        try:
            os.remove('test_model_defn_asset.zip')
        except:
            pass

    def test_03_get_details(self):

        details = self.client.model_definitions.get_details(TestModelDefn.model_defn_asset_id)
        print(details)
        self.assertTrue(TestModelDefn.model_defn_asset_id in str(details))

    def test_04_list(self):
        self.client.model_definitions.list()
    #
    # def test_06_revisions(self):
    #     meta_prop_script1_revision1 = {
    #         self.client.shiny.ConfigurationMetaNames.NAME: "Script revision 1",
    #         self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "Script revision 1"
    #     }
    #
    #     # Update meta and attachment for revision 1 creation
    #     model_defn_update = self.client.model_definitions.update(TestModelDefn.model_defn_asset_id,
    #                                               meta_prop_script1_revision1,
    #                                               file_path="artifacts/test1.py.zip")
    #     attachment_id = model_defn_update[u'metadata'][u'attachment_id']
    #
    #     # Create revision
    #     revision = self.client.model_definitions.create_revision(TestModelDefn.model_defn_asset_id)
    #
    #     self.assertTrue(revision[u'metadata'][u'revision_id'] and revision[u'metadata'][u'revision_id'] == 1)
    #
    #     new_attachment_id = revision[u'metadata'][u'attachment_id']
    #
    #     self.assertTrue(attachment_id != new_attachment_id)
    #
    #     # List revisions
    #     self.client.model_definitions.list_revisions(TestModelDefn.model_defn_asset_id)
    #
    #     try:
    #         os.remove('script1.zip')
    #     except:
    #         pass
    #
    #     # Download revision 1 attachment
    #     self.client.model_definitions.download(TestModelDefn.model_defn_asset_id,
    #                                 filename='script1.zip',
    #                                 rev_uid=1)
    #     try:
    #         os.remove('script1.zip')
    #     except:
    #         pass

    def test_05_delete_model_defn_asset(self):
        TestModelDefn.logger.info("Delete model definition")
        self.client.model_definitions.delete(TestModelDefn.model_defn_asset_id)


if __name__ == '__main__':
    unittest.main()



"""
**Warning**
In order to execute those tests correctly please make sure data is already placed
under the specified location /schema_name/table_name.
(You can easily do this by running the `test_autoai_project_iris_using_database_connection.py` before those tests).
"""


import unittest

from ibm_watson_machine_learning.tests.utils import is_cp4d
from ibm_watson_machine_learning.tests.autoai.abstract_tests_classes import\
    AbstractTestAutoAIConnectedAsset


@unittest.skipIf(not is_cp4d(), "Not supported on Cloud")
class TestAutoAIMSSQLServer(AbstractTestAutoAIConnectedAsset, unittest.TestCase):
    database_name = "sqlserver"
    schema_name = "connections"
    max_connection_nb = None


@unittest.skipIf(not is_cp4d(), "Not supported on Cloud")
class TestAutoAIDB2(AbstractTestAutoAIConnectedAsset, unittest.TestCase):
    database_name = "db2"
    schema_name = "CJB94327"
    table_name = "IRIS"
    max_connection_nb = 1


@unittest.skipIf(not is_cp4d(), "Not supported on Cloud")
class TestAutoAIPostgresSQL(AbstractTestAutoAIConnectedAsset, unittest.TestCase):
    database_name = "postgresql"
    schema_name = "public"
    max_connection_nb = 15

@unittest.skipIf(not is_cp4d(), "Not supported on Cloud")
@unittest.skip("The writing of training data is broken for now.")
class TestAutoAIMySQL(AbstractTestAutoAIConnectedAsset, unittest.TestCase):
    database_name = "mysql"
    schema_name = "mysql"


if __name__ == "__main__":
    unittest.main()



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from typing import List, Union

from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.enums import (
    TShirtSize, ClassificationAlgorithms, RegressionAlgorithms, ForecastingAlgorithms, PredictionType, Metrics,\
    Transformers, DataConnectionTypes, PipelineTypes, PositiveLabelClass, ClassificationAlgorithmsCP4D,
    RegressionAlgorithmsCP4D, ForecastingAlgorithmsCP4D, SamplingTypes)
from ibm_watson_machine_learning.utils.autoai.errors import LocalInstanceButRemoteParameter, MissingPositiveLabel, \
    NonForecastPredictionColumnMissing, ForecastPredictionColumnsMissing, ForecastingCannotBeRunAsLocalScenario, \
    TSNotSupported, ParamOutOfRange
from ibm_watson_machine_learning.utils.autoai.utils import check_dependencies_versions, \
    validate_additional_params_for_optimizer, validate_optimizer_enum_values
from ibm_watson_machine_learning.workspace import WorkSpace
from .engines import WMLEngine
from .optimizers import LocalAutoPipelines, RemoteAutoPipelines
from .runs import AutoPipelinesRuns, LocalAutoPipelinesRuns
from ..base_experiment.base_experiment import BaseExperiment

__all__ = [
    "AutoAI"
]


class AutoAI(BaseExperiment):
    """
    AutoAI class for pipeline models optimization automation.

    Parameters
    ----------
    wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance.

    project_id: str, optional
        ID of the Watson Studio project.

    space_id: str, optional
        ID of the Watson Studio Space.

    Example
    -------
    >>> from ibm_watson_machine_learning.experiment import AutoAI
    >>>
    >>> experiment = AutoAI(
    >>>        wml_credentials={
    >>>              "apikey": "...",
    >>>              "iam_apikey_description": "...",
    >>>              "iam_apikey_name": "...",
    >>>              "iam_role_crn": "...",
    >>>              "iam_serviceid_crn": "...",
    >>>              "instance_id": "...",
    >>>              "url": "https://us-south.ml.cloud.ibm.com"
    >>>            },
    >>>         project_id="...",
    >>>         space_id="...")
    """
    # note: initialization of AutoAI enums as class properties

    # note: Enums with estimators can be overwritten  in _init based on environment type (CPD or Cloud)
    ClassificationAlgorithms = ClassificationAlgorithms
    RegressionAlgorithms = RegressionAlgorithms
    ForecastingAlgorithms = ForecastingAlgorithms
    # end note
    TShirtSize = TShirtSize
    PredictionType = PredictionType
    Metrics = Metrics
    Transformers = Transformers
    DataConnectionTypes = DataConnectionTypes
    PipelineTypes = PipelineTypes

    def __init__(self,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None) -> None:
        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        if wml_credentials is None:
            self._workspace = None
            self.runs = LocalAutoPipelinesRuns()

        else:
            if isinstance(wml_credentials, WorkSpace):
                self._workspace = wml_credentials
            else:
                self._workspace = WorkSpace(wml_credentials=wml_credentials.copy(),
                                            project_id=project_id,
                                            space_id=space_id)

            self.project_id = self._workspace.project_id
            self.space_id = self._workspace.space_id
            self.runs = AutoPipelinesRuns(engine=WMLEngine(self._workspace))
            self.runs._workspace = self._workspace

        self._init_estimator_enums()

        self._20_class_limit_removal_test = False
        # --- end note

    def runs(self, *, filter: str) -> Union['AutoPipelinesRuns', 'LocalAutoPipelinesRuns']:
        """
        Get the historical runs but with WML Pipeline name filter (for remote scenario).
        Get the historical runs but with experiment name filter (for local scenario).

        Parameters
        ----------
        filter: str, required
            WML Pipeline name to filter the historical runs.
            or experiment name to filter the local historical runs.

        Returns
        -------
        AutoPipelinesRuns or LocalAutoPipelinesRuns

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(...)
        >>>
        >>> experiment.runs(filter='Test').list()
        """

        if self._workspace is None:
            return LocalAutoPipelinesRuns(filter=filter)

        else:
            return AutoPipelinesRuns(engine=WMLEngine(self._workspace.wml_client), filter=filter)

    def optimizer(self,
                  name: str,
                  *,
                  prediction_type: 'PredictionType',
                  prediction_column: str = None,
                  prediction_columns: List[str] = None,
                  timestamp_column_name: str = None,
                  scoring: 'Metrics' = None,
                  desc: str = None,
                  test_size: float = None, # deprecated
                  holdout_size: float = 0.1,
                  max_number_of_estimators: int = 2,
                  train_sample_rows_test_size: float = None,
                  include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms',
                                                      'ForecastingAlgorithms']] = None,
                  daub_include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']] = None, # deprecated
                  backtest_num: int = None,
                  lookback_window: int = None,
                  forecast_window: int = None,
                  backtest_gap_length: int = None,
                  cognito_transform_names: List['Transformers'] = None,
                  data_join_graph: 'DataJoinGraph' = None,
                  csv_separator: Union[List[str], str] = ',',
                  excel_sheet: Union[str, int] = 0,
                  encoding: str = 'utf-8',
                  positive_label: str = None,
                  data_join_only: bool = False,
                  drop_duplicates: bool = True,
                  # fairness_info: dict = None, #TODO: Uncomment to add fairness support
                  text_processing: bool = None,
                  word2vec_feature_number: int = None,
                  daub_give_priority_to_runtime: float = None,
                  # sampling_type: 'SamplingTypes' = None, #TODO: Uncomment to add sampling_type support
                  n_parallel_data_connections: int =None,
                  **kwargs) -> Union['RemoteAutoPipelines', 'LocalAutoPipelines']:
        """
        Initialize an AutoAi optimizer.

        Parameters
        ----------
        name: str, required
            Name for the AutoPipelines

        prediction_type: PredictionType, required
            Type of the prediction.

        prediction_column: str, required for multiclass, binary and regression prediction types
            Name of the target/label column

        prediction_columns: List[str], required for forecasting prediction type
            Names of the target/label columns.

        timestamp_column_name: str, optional, used only for forecasting prediction type
            Name of timestamp column for time series forecasting.

        scoring: Metrics, optional
            Type of the metric to optimize with. Not used for forecasting.

        desc: str, optional
            Description

        test_size: deprecated
            Use `holdout_size` instead.

        holdout_size: float, optional
            Percentage of the entire dataset to leave as a holdout. Default 0.1

        max_number_of_estimators: int, optional
            Maximum number (top-K ranked by DAUB model selection) of the selected algorithm, or estimator types,
            for example LGBMClassifierEstimator, XGBoostClassifierEstimator, or LogisticRegressionEstimator
            to use in pipeline composition.  The default is 2, where only the highest ranked by model
            selection algorithm type is used. (min 1, max 4)

        train_sample_rows_test_size: float, optional
            Training data sampling percentage

        daub_include_only_estimators: deprecated
            Use `include_only_estimators` instead.

        include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms', 'ForecastingAlgorithms']], optional
            List of estimators to include in computation process.
            See: AutoAI.ClassificationAlgorithms, AutoAI.RegressionAlgorithms or AutoAI.ForecastingAlgorithms

        backtest_num: int, optional
            Used for forecasting prediction type. Configure number of backtests. Default value: 4
            Value from range [0, 20]

        lookback_window: int, optional
            Used for forecasting prediction type. Configure length of lookback window. Default value: 10
            If set to -1 lookback window will be auto-detected

        forecast_window: int, optional
            Used for forecasting prediction type. Configure length of forecast window. Default value: 1
            Value from range [1, 60]

        backtest_gap_length: int, optional
            Used for forecasting prediction type. Configure gap between backtests. Default value: 0
            Value from range [0, data length / 4]

        cognito_transform_names: List['Transformers'], optional
            List of transformers to include in feature enginnering computation process.
            See: AutoAI.Transformers

        csv_separator: Union[List[str], str], optional
            The separator, or list of separators to try for separating
            columns in a CSV file.  Not used if the file_name is not a CSV file.
            Default is ','.

        excel_sheet: Union[str, int], optional
            Name or number of the excel sheet to use. Only use when xlsx file is an input.
            Default is 0.

        encoding: str, optional
            Encoding type for CSV training file.

        positive_label: str, optional
            The positive class to report when binary classification.
            When multiclass or regression, this will be ignored.

        t_shirt_size: TShirtSize, optional
            The size of the remote AutoAI POD instance (computing resources). Only applicable to a remote scenario.
            See: AutoAI.TShirtSize

        data_join_graph: DataJoinGraph, optional
            A graph object with definition of join structure for multiple input data sources.
            Data preprocess step for multiple files.

        data_join_only: bool, optional
            If True only preprocessing will be executed.

        drop_duplicates: bool, optional
            If True duplicated rows in data will be removed before further processing. Default is True.

        text_processing: bool, optional
            If True text processing will be enabled. Default is True. Applicable only on Cloud.

        word2vec_feature_number: int, optional
            Number of features which will be generated from text column. Will be applied only if `text_processing`
            is True. If `None` the default value will be taken.

        daub_give_priority_to_runtime: float, optional
            The importance of run time over score for pipelines ranking. Can take values between 0 and 5.
            If set to 0.0 only score is used. If set to 1 equally score and runtime are used.
            If set to value higher than 1 the runtime gets higher importance over score.

        n_parallel_data_connections: int, optional
            Number of maximum parallel connection to data source.
            Supported only for IBM Cloud Pak® for Data 4.0.1 and above.

        Returns
        -------
        RemoteAutoPipelines or LocalAutoPipelines, depends on how you initialize the AutoAI object.

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(...)
        >>>
        >>>
        >>> optimizer = experiment.optimizer(
        >>>        name="name of the optimizer.",
        >>>        prediction_type=AutoAI.PredictionType.BINARY,
        >>>        prediction_column="y",
        >>>        scoring=AutoAI.Metrics.ROC_AUC_SCORE,
        >>>        desc="Some description.",
        >>>        holdout_size=0.1,
        >>>        max_num_daub_ensembles=1,
        >>>        cognito_transform_names=[AutoAI.Transformers.SUM,AutoAI.Transformers.MAX],
        >>>        train_sample_rows_test_size=1,
        >>>        include_only_estimators=[AutoAI.ClassificationAlgorithms.LGBM, AutoAI.ClassificationAlgorithms.XGB],
        >>>        t_shirt_size=AutoAI.TShirtSize.L
        >>>    )
        >>>
        >>> optimizer = experiment.optimizer(
        >>>        name="name of the optimizer.",
        >>>        prediction_type=AutoAI.PredictionType.MULTICLASS,
        >>>        prediction_column="y",
        >>>        scoring=AutoAI.Metrics.ROC_AUC_SCORE,
        >>>        desc="Some description.",
        >>>    )
        """

        # # TODO: Pass the fairness info description and example when adding fairness support and
        #
        # """
        # fairness_info: dict, optional
        #     Dictionary that  specifies metadata needed for measuring fairness. It contains two key values:
        #     favorable_labels and protected_attributes. The favorable_labels attribute indicates that when the class
        #     column contains one of the value from list, that is considered a positive outcome.
        #     A protected attribute is a feature that partitions the population into groups whose outcome should have parity.
        #     If passed fairness metric will be calculated.
        # >>> fairness_info = {
        # >>>        "protected_attributes": [
        # >>>            {"feature": "Sex", "privileged_groups": ['female']},
        # >>>             {"feature": "Age", "privileged_groups": [[20, 40], [60, 90]]}
        # >>>        ],
        # >>>        "favorable_labels": ["No Risk"]
        # >>>    }
        # """

        # TODO: Pass the sampling_type doc description, when adding support for that
        # sampling_type: str, optional
        #     Type of sampling data for training. One of SamplingTypes enum values.
        #     Default is SamplingTypes.FIRST_N_RECORDS.
        #     Supported only for IBM Cloud Pak® for Data 4.0.1 and above.

        if prediction_type == PredictionType.FORECASTING and self._workspace.wml_client.ICP and \
                (self._workspace.wml_client.wml_credentials['version'].startswith('2.5') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.0') or \
                        self._workspace.wml_client.wml_credentials['version'].startswith('3.5')):
            raise TSNotSupported()

        if prediction_type == PredictionType.FORECASTING:
            if prediction_column is not None or prediction_columns is None:
                raise ForecastPredictionColumnsMissing()
        else:
            if prediction_column is None or prediction_columns is not None:
                raise NonForecastPredictionColumnMissing(prediction_type)

        if test_size:
            print('Note: Using `test_size` is deprecated. Use `holdout_size` instead.')
            if not holdout_size:
                holdout_size = test_size
            test_size = None

        if daub_include_only_estimators:
            print('Note: Using `daub_include_only_estimators` is deprecated. Use `include_only_estimators` instead.')
            if not include_only_estimators:
                include_only_estimators = daub_include_only_estimators
            daub_include_only_estimators = None

        validate_optimizer_enum_values(
            prediction_type=prediction_type,
            daub_include_only_estimators=daub_include_only_estimators,
            include_only_estimators=include_only_estimators,
            cognito_transform_names=cognito_transform_names,
            scoring=scoring,
            t_shirt_size=kwargs.get("t_shirt_size", TShirtSize.M),
            is_cpd=self._workspace.wml_client.ICP
        )

        if daub_give_priority_to_runtime is not None:
            if daub_give_priority_to_runtime < 0.0 or daub_give_priority_to_runtime > 5.0:
                raise ParamOutOfRange('daub_give_priority_to_runtime', daub_give_priority_to_runtime, 0.0, 5.0)

        if data_join_graph:
            data_join_graph.problem_type = prediction_type
            data_join_graph.target_column = prediction_column

        if (prediction_type == PredictionType.BINARY and scoring in vars(PositiveLabelClass).values()
                and positive_label is None):
            raise MissingPositiveLabel(scoring, reason=f"\"{scoring}\" needs a \"positive_label\" "
                                                       f"parameter to be defined when used with binary classification.")

        if self._workspace is None and kwargs.get('t_shirt_size'):
            raise LocalInstanceButRemoteParameter(
                "t_shirt_size",
                reason="During LocalOptimizer initialization, \"t_shirt_size\" parameter was provided. "
                       "\"t_shirt_size\" parameter is only applicable to the RemoteOptimizer instance."
            )
        elif self._workspace is None:
            if prediction_type == PredictionType.FORECASTING:
                raise ForecastingCannotBeRunAsLocalScenario()

            reduced_kwargs = copy.copy(kwargs)

            for n in ['_force_local_scenario']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            return LocalAutoPipelines(
                name=name,
                prediction_type='classification' if prediction_type in ['binary', 'multiclass'] else prediction_type,
                prediction_column=prediction_column,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                cognito_transform_names=cognito_transform_names,
                positive_label=positive_label,
                _force_local_scenario=kwargs.get('_force_local_scenario', False),
                **reduced_kwargs
            )

        else:
            reduced_kwargs = copy.copy(kwargs)

            for n in ['t_shirt_size', 'notebooks', 'autoai_pod_version', 'obm_pod_version']:
                if n in reduced_kwargs:
                    del reduced_kwargs[n]

            validate_additional_params_for_optimizer(reduced_kwargs)

            engine = WMLEngine(self._workspace)

            if self._20_class_limit_removal_test:
                engine._20_class_limit_removal_test = True

            optimizer = RemoteAutoPipelines(
                name=name,
                prediction_type=prediction_type,
                prediction_column=prediction_column,
                prediction_columns=prediction_columns,
                timestamp_column_name=timestamp_column_name,
                scoring=scoring,
                desc=desc,
                holdout_size=holdout_size,
                max_num_daub_ensembles=max_number_of_estimators,
                t_shirt_size=self._workspace.restrict_pod_size(t_shirt_size=kwargs.get(
                    't_shirt_size', TShirtSize.M if self._workspace.wml_client.ICP else TShirtSize.L)
                ),
                train_sample_rows_test_size=train_sample_rows_test_size,
                include_only_estimators=include_only_estimators,
                backtest_num=backtest_num,
                lookback_window=lookback_window,
                forecast_window=forecast_window,
                backtest_gap_length=backtest_gap_length,
                cognito_transform_names=cognito_transform_names,
                data_join_graph=data_join_graph,
                drop_duplicates=drop_duplicates,
                text_processing=text_processing,
                word2vec_feature_number=word2vec_feature_number,
                csv_separator=csv_separator,
                excel_sheet=excel_sheet,
                encoding=encoding,
                positive_label=positive_label,
                data_join_only=data_join_only,
                engine=engine,
                daub_give_priority_to_runtime=daub_give_priority_to_runtime,
                notebooks=kwargs.get('notebooks', True),
                autoai_pod_version=kwargs.get('autoai_pod_version', None),
                obm_pod_version=kwargs.get('obm_pod_version', None),
                # fairness_info=fairness_info, #TODO: Uncomment to add fairness support
                # sampling_type=sampling_type, #TODO: Uncomment to add sampling_type support
                n_parallel_data_connections=n_parallel_data_connections,
                **reduced_kwargs
            )
            optimizer._workspace = self._workspace
            return optimizer

    def _init_estimator_enums(self):
        if self._workspace and self._workspace.wml_client.ICP:
            self.ClassificationAlgorithms = ClassificationAlgorithmsCP4D
            self.RegressionAlgorithms = RegressionAlgorithmsCP4D
            self.ForecastingAlgorithms = ForecastingAlgorithmsCP4D
        else:
            self.ClassificationAlgorithms = ClassificationAlgorithms
            self.RegressionAlgorithms = RegressionAlgorithms
            self.ForecastingAlgorithms = ForecastingAlgorithms



import unittest

from ibm_watson_machine_learning.helpers import DataConnection, S3Connection, S3Location, AssetLocation
from ibm_watson_machine_learning.utils.autoai.utils import prepare_cos_client
from ibm_boto3 import resource, client
from ibm_botocore.client import Config


cos_resource = resource(
    service_name='s3',
    ibm_api_key_id='...',
    ibm_auth_endpoint='...',
    config=Config(signature_version="oauth"),
    endpoint_url='https://s3-api.us-geo.objectstorage.softlayer.net'
)


class TestPrepareCOSClient(unittest.TestCase):
    def test_01_check_clients_types(self):
        training_data_reference = [
            DataConnection(
                connection=S3Connection(
                    access_key_id='...',
                    secret_access_key='...',
                    endpoint_url='https://s3-api.us-geo.objectstorage.softlayer.net'
                ),
                location=S3Location(
                    bucket='some_bucket',
                    path='some_path'
                )
            ),
            DataConnection(
                connection=S3Connection(
                    api_key='...',
                    auth_endpoint='...',
                    endpoint_url='https://s3-api.us-geo.objectstorage.softlayer.net'
                ),
                location=S3Location(
                    bucket='some_bucket_2',
                    path='some_path_2'
                )
            )
        ]

        training_result_reference = DataConnection(
            connection=S3Connection(
                api_key='...',
                auth_endpoint='...',
                endpoint_url='https://s3-api.us-geo.objectstorage.softlayer.net'
            ),
            location=S3Location(
                bucket='some_bucket_2',
                path='some_path_2'
            )
        )

        data_clients, result_client = prepare_cos_client(training_data_references=training_data_reference,
                                                         training_result_reference=training_result_reference)

        self.assertIsInstance(data_clients, list, msg="data_clients is not a list!")
        self.assertIsInstance(data_clients[0][0], DataConnection,
                              msg='First variable from data_clients[0] tuple should be always a DataConnection')
        self.assertIsInstance(data_clients[1][0], DataConnection,
                              msg='First variable from data_clients[1] tuple should be always a DataConnection')
        self.assertIsInstance(result_client[0], DataConnection,
                              msg='First variable from result_client tuple should be always a DataConnection')
        self.assertEqual(str(type(data_clients[0][1])), str(type(cos_resource)),
                         msg="Wrong COS client type in [0] of data_clients")
        self.assertEqual(str(type(data_clients[1][1])), str(type(cos_resource)),
                         msg="Wrong COS client type in [1] of data_clients")
        self.assertEqual(str(type(result_client[1])), str(type(cos_resource)),
                         msg="Wrong COS client type in result_client")

    def test_02_wrong_connections(self):
        training_data_reference = [
            DataConnection(
                location=AssetLocation(
                    asset_id='...'
                )
            ),
            DataConnection(
                location=AssetLocation(
                    asset_id='...'
                )
            )
        ]

        training_result_reference = DataConnection(
            location=AssetLocation(
                asset_id='...'
            )
        )

        data_clients, result_client = prepare_cos_client(training_data_references=training_data_reference,
                                                         training_result_reference=training_result_reference)

        self.assertIsInstance(data_clients, list, msg="data_clients is not a list!")
        self.assertEqual(data_clients, [], msg="data_clients should be an empty list!")
        self.assertIsNone(result_client, msg="result_client should be None")


if __name__ == '__main__':
    unittest.main()



import io
import os
import abc
import contextlib

import ibm_boto3

from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.tests.utils import bucket_exists, create_bucket
from ibm_watson_machine_learning.tests.base.abstract.abstract_client_test import AbstractClientTest


class AbstractDeepLearningTest(AbstractClientTest, abc.ABC):
    """
    Class implementing interface of test checking a scenario
    of deep learning training.
    """
    model_definition_name = None
    training_name = None
    training_description = None
    software_specification_name = None
    execution_command = None

    training_id = None
    training_props = None
    data_location = None
    data_cos_path = None

    # If Test case is composed of multiple models pass the paths inside the \
    # List object [even if one model is required it should be passed wrapped in List].
    model_paths = None
    model_definition_ids = []

    SPACE_ONLY = False

    def create_model_definition_payload(self):
        """
        Creates payload for model definition.
        """
        return {
            self.wml_client.model_definitions.ConfigurationMetaNames.NAME: self.model_definition_name,
            self.wml_client.model_definitions.ConfigurationMetaNames.COMMAND: self.execution_command,
            self.wml_client.model_definitions.ConfigurationMetaNames.PLATFORM: {"name": "python", "versions": ["3.8"]},
            self.wml_client.model_definitions.ConfigurationMetaNames.VERSION: "2.0",
        }

    def create_training_payload(self):
        """
        Creates payload for training creation.
        """
        return {
            self.wml_client.training.ConfigurationMetaNames.NAME: self.training_name,
            self.wml_client.training.ConfigurationMetaNames.DESCRIPTION: self.training_description,
            self.wml_client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES: [
                {
                    "type": "connection_asset",
                    "connection": {
                        "id": self.connection_id,
                    },
                    "location": {
                        "bucket": self.bucket_name,
                        "file_name": "."
                    }
                }
            ],
            self.wml_client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE:  {
                "type": "connection_asset",
                "connection": {
                    "id": self.connection_id,
                },
                "location": {
                    "bucket": self.bucket_name,
                    "file_name": "."
                },
            },
            self.wml_client.training.ConfigurationMetaNames.MODEL_DEFINITION: {
                "id": self.definition_id,
                "hardware_spec": {
                  "name": "K80",
                  "nodes": 1
                },
                "software_spec": {
                  "name": self.software_specification_name
                },
            }
        }

    def test_01_prepare_COS_instance(self):
        AbstractDeepLearningTest.cos_resource = ibm_boto3.resource(
            service_name="s3",
            endpoint_url=self.cos_endpoint,
            aws_access_key_id=self.cos_credentials['cos_hmac_keys']['access_key_id'],
            aws_secret_access_key=self.cos_credentials['cos_hmac_keys']['secret_access_key']
        )
        # Prepare bucket
        if not bucket_exists(self.cos_resource, self.bucket_name):
            AbstractDeepLearningTest.bucket_name = create_bucket(self.cos_resource, self.bucket_name)

            self.assertIsNotNone(self.bucket_name)
            self.assertTrue(bucket_exists(self.cos_resource, AbstractDeepLearningTest.bucket_name))

        if os.path.isfile(self.data_location):
            self.cos_resource.Bucket(self.bucket_name).upload_file(
                self.data_location,
                self.data_cos_path
            )
        else:
            for file in os.listdir(self.data_location):
                self.cos_resource.Bucket(self.bucket_name).upload_file(
                    os.path.join(self.data_location, file),
                    self.data_cos_path
                )

    def test_02_prepare_connection_to_COS(self):
        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name('bluemixcloudobjectstorage'),
            'name': 'Connection to COS for tests',
            'properties': {
                'bucket': self.bucket_name,
                'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                'iam_url': self.wml_client.service_instance._href_definitions.get_iam_token_url(),
                'url': self.cos_endpoint
            }
        })

        AbstractDeepLearningTest.connection_id = self.wml_client.connections.get_uid(connection_details)
        self.assertIsInstance(self.connection_id, str)

    def test_03_create_model_definition(self):
        for model_path in self.model_paths:
            meta_props = self.create_model_definition_payload()
            model_definitions_details = self.wml_client.model_definitions.store(
                model_path, meta_props
            )
            model_definition_id = self.wml_client.model_definitions.get_id(
                model_definitions_details
            )
            self.assertIsNotNone(model_definition_id)
            self.model_definition_ids.append(model_definition_id)

        self.assertEqual(len(self.model_definition_ids), len(self.model_paths))

    def test_04_get_model_definition_details(self):
        for id_ in self.model_definition_ids:
            details = self.wml_client.model_definitions.get_details(id_)
            self.assertIsNotNone(details)
            self.assertEqual(self.model_definition_name, details["metadata"]["name"])

    def test_05_list_definitions(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.model_definitions.list()
            definitions_list = buf.getvalue()

        self.assertTrue(all(id_ in definitions_list for id_ in self.model_definition_ids))

    def test_06_create_definition_revision(self):
        AbstractDeepLearningTest.definition_id = self.model_definition_ids[0]
        revision = self.wml_client.model_definitions.create_revision(self.definition_id)
        AbstractDeepLearningTest.rev_id = revision['metadata'].get('revision_id')
        self.assertIsNotNone(self.rev_id)

    def test_07_list_definition_revisions(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.model_definitions.list_revisions(self.definition_id)
            rev_list = buf.getvalue()

        self.assertIn(str(self.rev_id), rev_list)

    def test_08_update_model_definition(self):
        details = self.wml_client.model_definitions.update(
            self.definition_id,
            {
                self.wml_client.model_definitions.ConfigurationMetaNames.NAME: "Updated model definition"
            }
        )
        self.assertIsNotNone(details)


    def test_09_download_model_definition_for_given_rev_id(self):
        definition_path = 'model.tar.gz'
        with contextlib.suppress(FileNotFoundError):
            os.remove(definition_path)

            self.wml_client.model_definitions.download(
                model_definition_uid=self.definition_id, filename=definition_path, rev_id=self.rev_id
            )

            os.remove(definition_path)

    def test_10_run_traininng(self):
        meta_props = self.create_training_payload()
        training_details = self.wml_client.training.run(meta_props, asynchronous=False)

        AbstractDeepLearningTest.training_id = self.wml_client.training.get_id(training_details)
        self.assertIsNotNone(self.training_id)

        state = self.wml_client.training.get_status(self.training_id).get('state')
        self.assertEqual(state, 'completed')

    def test_11a_get_training_details(self):
        details = self.wml_client.training.get_details(self.training_id)
        self.assertIsNotNone(details)
        self.assertEqual(self.training_name, details["metadata"]["name"])

    def test_11b_get_trainings_details(self):
        data = self.wml_client.training.get_details(asynchronous=False, get_all=True, limit=10)
        for synch_entity, asynch_entity in zip(
                data, self.wml_client.training.get_details(asynchronous=False, get_all=True, limit=10)):
            print(asynch_entity)
            self.assertIsNotNone(asynch_entity)
            self.assertEqual(synch_entity, asynch_entity)

    def test_12_get_training_list(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.training.list(limit=200)
            training_list = buf.getvalue()

        # TODO check listing
        # self.assertIn(self.training_id, training_list)

    def test_13_hard_delete_training(self):
        self.wml_client.training.cancel(self.training_id, hard_delete=True)

    def test_14_delete_model_definitions(self):
        for id_ in self.model_definition_ids:
            self.wml_client.model_definitions.delete(id_)

        with self.assertRaises(WMLClientError):
            self.wml_client.model_definitions.get_details(self.definition_id)


class AbstractDeepLearningExperimentTest(AbstractDeepLearningTest, abc.ABC):
    """
    Class implementing interface of test checking a scenario
    of deep learning training using experiment.
    """
    experiment_name = None

    def create_experiment_payload(self):
        """
        Creates payload for experiment.
        """
        return {
            self.wml_client.repository.ExperimentMetaNames.NAME: self.experiment_name,
            self.wml_client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                {
                    "model_definition": {
                        "id": self.definition_id,
                        "hardware_spec": {
                            "name": "K80",
                            "nodes": 1
                        },
                        "software_spec": {
                            "name": self.software_specification_name
                        }
                    }
                }
            ]
        }

    def create_training_payload(self):
        """
        Overrides payload for training creation.
        It's using experiment instead of model definition.
        """
        return {
            self.wml_client.training.ConfigurationMetaNames.NAME: self.training_name,
            self.wml_client.training.ConfigurationMetaNames.DESCRIPTION: self.training_description,
            self.wml_client.training.ConfigurationMetaNames.TRAINING_DATA_REFERENCES: [
                {
                    "type": "connection_asset",
                    "connection": {
                        "id": self.connection_id,
                    },
                    "location": {
                        "bucket": self.bucket_name,
                        "file_name": "."
                    }
                }
            ],
            self.wml_client.training.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE:  {
                "type": "connection_asset",
                "connection": {
                    "id": self.connection_id,
                },
                "location": {
                    "bucket": self.bucket_name,
                    "file_name": "."
                },
            },
            self.wml_client.training.ConfigurationMetaNames.EXPERIMENT: {
                "id": self.experiment_id
            }
        }

    def test_09a_store_experiment(self):
        meta_props = self.create_experiment_payload()
        experiment_details = self.wml_client.experiments.store(meta_props)
        AbstractDeepLearningExperimentTest.experiment_id = \
            self.wml_client.experiments.get_id(experiment_details)

    def test_09b_get_experiment_details(self):
        details = self.wml_client.experiments.get_details(self.experiment_id)
        self.assertIsNotNone(details)
        self.assertEqual(self.experiment_name, details["metadata"]["name"])

    def test_09c_list_experiments(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.experiments.list()
            experiments_list = buf.getvalue()

        self.assertIn(self.experiment_id, experiments_list)

    def test_09d_create_experiment_revision(self):
        revision = self.wml_client.experiments.create_revision(self.experiment_id)
        AbstractDeepLearningTest.rev_id = revision['metadata'].get('rev')
        self.assertIsNotNone(self.rev_id)

    def test_09e_list_experiment_revisions(self):
        with io.StringIO() as buf, contextlib.redirect_stdout(buf):
            self.wml_client.experiments.list_revisions(self.experiment_id)
            rev_list = buf.getvalue()

        self.assertIn(self.rev_id, rev_list)

    def test_09f_update_experiment(self):
        updated_name = "Updated DL experiment"
        details = self.wml_client.experiments.update(
            self.experiment_id,
            {
                self.wml_client.experiments.ConfigurationMetaNames.NAME: updated_name
            }
        )

        self.assertIsNotNone(details)
        self.assertEqual(updated_name, details["metadata"]["name"])



import unittest
import traceback

from lale.operators import TrainablePipeline
from lale.lib.lale import Hyperopt
from lale.operators import TrainedPipeline
from lale import wrap_imported_operators
from lale.helpers import import_from_sklearn_pipeline

from sklearn.linear_model import LogisticRegression as LR
from sklearn.tree import DecisionTreeClassifier as Tree
from sklearn.neighbors import KNeighborsClassifier as KNN

# from lale.lib.sklearn import KNeighborsClassifier, DecisionTreeClassifier, LogisticRegression



from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService

from ibm_watson_machine_learning.experiment.autoai.autoai import PipelineTypes
from ibm_watson_machine_learning.tests.utils import get_wml_credentials, is_cp4d



class TestLale(unittest.TestCase):
    """
    Test Case Description:

    1. Get historical optimizer described by run_id or if run_id is None -> get last autoai optimizer.
    2. Get all generated pipelines as lale type and predict.
    3. Pipeline refinery with lale:
        * Remove last Pipeline stage.
        * Add three estimator choices as a last Pipeline stage.
        * Use HyperOpt to find best Pipeline (best last step estimator)
    4. Deploy lale pipeline
    """

    wml_credentials = None
    project_id = None
    space_id = None

    experiment = None
    lale_pipeline = None
    new_pipeline = None
    prefix = None
    hist_opt = None
    hyperopt_pipelines = None
    new_sklearn_pipeline = None
    service: 'WebService' = None
    
    X_data = None
    y_data = None

    run_id = None

    optimizer_name = None


    @classmethod
    def setUp(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """
        wrap_imported_operators()
        cls.wml_credentials = get_wml_credentials()

    def test_01_initialize_AutoAI_experiment__pass_credentials__object_initialized(self):
        if is_cp4d():
            TestLale.experiment = AutoAI(wml_credentials=self.wml_credentials.copy(),
                                                 project_id=self.project_id,
                                                 space_id=self.space_id)
        else:
            TestLale.experiment = AutoAI(wml_credentials=self.wml_credentials)

        self.assertIsInstance(self.experiment, AutoAI, msg="Experiment is not of type AutoAI.")

    def test_02_get_run_id(self):
        if not self.run_id:
            # IF run id not set get last complete autoai run.
            runs_df = self.experiment.runs(filter=self.optimizer_name).list()

            self.assertGreater(len(runs_df), 0,
                               msg="No runs found for the WML instance. Please run any autoai experiment (or test) with the wml instance before rerun the test")

            for run in runs_df.values:
                if run[2] == 'completed':
                    TestLale.run_id = run[1]
                    break
            self.assertIsNotNone(TestLale.run_id,
                               msg="No completed runs found for the WML instance. Please run any autoai experiment (or test) with the wml instance before rerun the test")

        print("AutoAI Run Id: ", TestLale.run_id)

        self.assertIsNotNone(TestLale.run_id)
        self.assertIsInstance(TestLale.run_id, str)

    def test_03_get_optimizer(self):
        TestLale.hist_opt = self.experiment.runs.get_optimizer(run_id=self.run_id)
        print(f"Optimizer params: {TestLale.hist_opt.get_params()}")

        data = TestLale.hist_opt.get_data_connections()[0].read(with_holdout_split=False)
        TestLale.X_data = data.drop(TestLale.hist_opt.get_params()['prediction_column'], axis=1)
        TestLale.y_data = data[TestLale.hist_opt.get_params()['prediction_column']]

        self.assertIsNotNone(TestLale.X_data, msg="Source data set is None/empty")
        self.assertGreater(len(TestLale.X_data), 0, msg="Source data set is empty")

        print(TestLale.X_data.head(3))

    def test_04_get_default_pipeline_astype_lale(self):
        TestLale.lale_pipeline = self.hist_opt.get_pipeline()

        print(f"Fetched pipeline type: {type(self.lale_pipeline)}")
        self.assertIsInstance(self.lale_pipeline, TrainablePipeline)

        predictions = self.lale_pipeline.predict(
            X=self.X_data.values[:5])

        print(predictions)
        self.assertGreater(len(predictions), 0,  msg="Returned prediction are empty")

    def test_05_get_all_pipelines_as_lale(self):
        summary = self.hist_opt.summary()
        print(summary)
        failed_pipelines = []
        for pipeline_name in summary.reset_index()['Pipeline Name']:
            print(f"Getting pipeline: {pipeline_name}")
            lale_pipeline = None
            try:
                lale_pipeline = self.hist_opt.get_pipeline(pipeline_name=pipeline_name)
                self.assertIsInstance(lale_pipeline, TrainablePipeline)
                predictions = lale_pipeline.predict(
                    X=self.X_data.values[:1])
                print(predictions)
                self.assertGreater(len(predictions), 0, msg=f"Returned prediction for {pipeline_name} are empty")
            except:
                print(f"Failure: {pipeline_name}")
                failed_pipelines.append(pipeline_name)
                traceback.print_exc()

            if not TestLale.lale_pipeline:
                TestLale.lale_pipeline = lale_pipeline
                print(f"{pipeline_name} loaded for next test cases")

        self.assertEqual(len(failed_pipelines), 0, msg=f"Some pipelines failed. Full list: {failed_pipelines}")

    def test_06_pipeline_to_script__lale__pretty_print(self):
        pipeline_code = self.lale_pipeline.pretty_print()
        try:
            exec(pipeline_code)

        except Exception as exception:
            self.assertIsNone(exception, msg="Pretty print from lale pipeline was not successful")

    ################################
    # Pipeline refinery with lale  #
    ################################

    def test_07_remove_last_freeze_trainable_prefix_returned(self):
        TestLale.prefix = self.lale_pipeline.remove_last().freeze_trainable()
        self.assertIsInstance(self.prefix, TrainablePipeline,
                              msg="Prefix pipeline is not of TrainablePipeline instance.")

    def test_08_add_estimator(self):
        TestLale.new_pipeline = self.prefix >> (LR | Tree | KNN)

    def test_09_hyperopt_fit_new_pipepiline(self):
        hyperopt = Hyperopt(estimator=TestLale.new_pipeline, cv=3, max_evals=2)
        try:
            TestLale.hyperopt_pipelines = hyperopt.fit(self.X_data.values, self.y_data.values)
        except ValueError as e:
            print(f"ValueError message: {e}")
            traceback.print_exc()
            hyperopt_results = hyperopt._impl._trials.results
            print(hyperopt_results)
            self.assertIsNone(e, msg="hyperopt fit was not successful")

    def test_10_get_pipeline_from_hyperopt(self):
        from sklearn.pipeline import Pipeline
        new_pipeline_model = TestLale.hyperopt_pipelines.get_pipeline()
        print(f"Hyperopt_pipeline_model is type: {type(new_pipeline_model)}")
        TestLale.new_sklearn_pipeline = new_pipeline_model.export_to_sklearn_pipeline()
        self.assertIsInstance(TestLale.new_sklearn_pipeline, Pipeline,
                              msg=f"Incorect Sklearn Pipeline type after conversion. Current: {type(TestLale.new_sklearn_pipeline)}")

    def test_11_predict_refined_pipeline(self):
        predictions = TestLale.new_sklearn_pipeline.predict(
            X=self.X_data.values[:1])
        print(predictions)
        self.assertGreater(len(predictions), 0, msg=f"Returned prediction for refined pipeline are empty")

    ############################
    # Deploy refined pipeline  #
    ############################

    def test_20_deployment_setup_and_preparation(self):
        if is_cp4d():
            TestLale.service = WebService(source_wml_credentials=self.wml_credentials.copy(),
                                          source_project_id=self.project_id,
                                          source_space_id=self.space_id)
        else:
            TestLale.service = WebService(source_wml_credentials=self.wml_credentials)

    def test_21__deploy__deploy_best_computed_pipeline_from_autoai_on_wml(self):
        deployment_name = TestLale.hist_opt.get_params()['name'] + " Deployment"
        model_to_deploy = import_from_sklearn_pipeline(self.new_sklearn_pipeline) \
            if self.new_sklearn_pipeline else self.lale_pipeline
        self.service.create(
            experiment_run_id=self.run_id,
            model=model_to_deploy,
            deployment_name=deployment_name)

    def test_22_score_deployed_model(self):
        nb_records = 5
        predictions = self.service.score(payload=self.X_data[:nb_records])
        print(predictions)
        self.assertIsNotNone(predictions)
        self.assertEqual(len(predictions['predictions'][0]['values']), nb_records)

    def test_23_list_deployments(self):
        self.service.list()
        params = self.service.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_24_delete_deployment(self):
        print("Delete current deployment: {}".format(self.service.deployment_id))
        self.service.delete()


if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys
from os.path import join as path_join

# SPARK_HOME_PATH = os.environ['SPARK_HOME']
# PYSPARK_PATH = str(SPARK_HOME_PATH) + "/python/"
# sys.path.insert(1, path_join(PYSPARK_PATH))

import logging
from preparation_and_cleaning import *
from models_preparation import *


class TestWMLClientWithSpark(unittest.TestCase):
    deployment_uid = None
    space_uid = None
    space_href = None
    model_uid = None
    scoring_url = None
    space_id = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithSpark.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.model_name = "SparkMLlibFromObjectLocal Model"
        self.deployment_name = "Test deployment"
        self.space_id = None

    def test_01_set_space(self):
        self.space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "space_with_spark_model_sw"})

        TestWMLClientWithSpark.space_id = self.client.spaces.get_uid(self.space)
        self.client.set.default_space(TestWMLClientWithSpark.space_id)
        self.assertTrue("SUCCESS" in self.client.set.default_space(TestWMLClientWithSpark.space_id))


    def test_02_publish_model(self):
        TestWMLClientWithSpark.logger.info("Creating spark model ...")

        model_data = create_spark_mllib_model_data()

        TestWMLClientWithSpark.logger.info("Publishing spark model ...")

        spark_sw_spec_uid = self.client.software_specifications.get_uid_by_name("spark-mllib_2.4")


        self.client.repository.ModelMetaNames.show()

        model_props = {self.client.repository.ModelMetaNames.NAME: "Spark 2.4 model",
            self.client.repository.ModelMetaNames.TYPE: "mllib_2.4",
            self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: spark_sw_spec_uid,
                       }

        published_model = self.client.repository.store_model(model=model_data['model'], meta_props=model_props, training_data=model_data['training_data'], pipeline=model_data['pipeline'])
        TestWMLClientWithSpark.model_uid = self.client.repository.get_model_uid(published_model)
        TestWMLClientWithSpark.logger.info("Published model ID:" + str(TestWMLClientWithSpark.model_uid))
        self.assertIsNotNone(TestWMLClientWithSpark.model_uid)

    def test_04_get_details(self):
        TestWMLClientWithSpark.logger.info("Get details")
        details = self.client.repository.get_details(self.model_uid)
        print(details)
        TestWMLClientWithSpark.logger.debug("Model details: " + str(details))
        self.assertTrue("Spark" in str(details))

    def test_05_create_deployment(self):
        TestWMLClientWithSpark.logger.info("Create deployment")
        deployment = self.client.deployments.create(self.model_uid, meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test deployment",self.client.deployments.ConfigurationMetaNames.ONLINE: {}})


        TestWMLClientWithSpark.logger.info("model_uid: " + self.model_uid)
        TestWMLClientWithSpark.logger.debug("Online deployment: " + str(deployment))
        TestWMLClientWithSpark.scoring_url = self.client.deployments.get_scoring_href(deployment)
        TestWMLClientWithSpark.logger.debug("Scoring url: {}".format(TestWMLClientWithSpark.scoring_url))
        TestWMLClientWithSpark.deployment_uid = self.client.deployments.get_uid(deployment)
        TestWMLClientWithSpark.logger.debug("Deployment uid: {}".format(TestWMLClientWithSpark.deployment_uid))
        self.assertTrue("online" in str(deployment))

    def test_06_get_deployment_details(self):
        TestWMLClientWithSpark.logger.info("Get deployment details")
        deployment_details = self.client.deployments.get_details(TestWMLClientWithSpark.deployment_uid)
        print(deployment_details)
        TestWMLClientWithSpark.logger.debug("Deployment details: {}".format(deployment_details))
        self.assertTrue(self.deployment_name in str(deployment_details))

    def test_07_score(self):
        TestWMLClientWithSpark.logger.info("Score the model")
        scoring_data = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA: [
                {
                    "fields": ["GENDER","AGE","MARITAL_STATUS","PROFESSION"],
                    "values": [["M",23,"Single","Student"],["M",55,"Single","Executive"]]
                }
            ]
        }
        predictions = self.client.deployments.score(TestWMLClientWithSpark.deployment_uid, scoring_data)
        print("Predictions: {}".format(predictions))
        self.assertTrue("prediction" in str(predictions))

    def test_08_delete_deployment(self):
        TestWMLClientWithSpark.logger.info("Delete deployment")
        self.client.deployments.delete(TestWMLClientWithSpark.deployment_uid)

    def test_09_delete_model(self):
        TestWMLClientWithSpark.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithSpark.model_uid)

    def test_10_delete_space(self):
        TestWMLClientWithSpark.logger.info("Delete space")
        self.client.spaces.delete(TestWMLClientWithSpark.space_id)


if __name__ == '__main__':
    unittest.main()
